import typing
from abc import abstractmethod, ABC
from collections import OrderedDict
from typing import Union, Callable

import v0.relationalai.semantics as qb
from v0.relationalai.semantics import define, where
from v0.relationalai.semantics.internal.internal import ConceptConstruct, Ref, RelationshipRef, ConceptFilter
from v0.relationalai.early_access.dsl.bindings.common import Binding, BindableAttribute, IdentifierConceptBinding, \
    SubtypeConceptBinding, BindableTable, ReferentConceptBinding
from v0.relationalai.early_access.dsl.codegen.helpers import reference_entity, construct_entity
from v0.relationalai.early_access.dsl.orm.constraints import Unique
from v0.relationalai.early_access.dsl.orm.relationships import Role
from v0.relationalai.semantics.metamodel.util import OrderedSet


def _get_map_madlib_and_name(name_prefix: str, value_concept: qb.Concept) -> tuple[str, str]:
    """
    Generates a name for the map based on the value concept.
    """
    name = f'{name_prefix}_row_to_{value_concept}'
    madlib = f'{name} {{row_id:RowId}} {{val:{value_concept}}}'
    return madlib, name


class GeneratedRelation(qb.Relationship):
    """
    A class representing a relation generated from a model.
    """
    def __init__(self, madlib, model, name):
        super().__init__(madlib, model=model.qb_model(), short_name=name)

    def __repr__(self):
        return f'GeneratedRelation({self._name})'


class InternallyGeneratedRelation(GeneratedRelation):
    """
    A class representing a relation generated by analyzing the model.
    """
    def __init__(self, madlib, dsl_model, name):
        super().__init__(madlib, dsl_model, name)


class AbstractMap(GeneratedRelation):
    """
    A class representing an abstract map.
    """
    def __init__(self, madlib, dsl_model, name):
        super().__init__(madlib, dsl_model, name)

    def value_player(self) -> qb.Concept:
        """
        Returns the value player of the map.
        This is the concept that the map maps to.
        """
        raise NotImplementedError()

    def formula(self):
        source_ref = self.table()
        rez_val_ref = self.value_player().ref()
        return rez_val_ref, [self(source_ref, rez_val_ref)]


class RoleMap(AbstractMap):
    """
    A class representing a base role map relation.
    """

    def __init__(self, madlib, model, name, role):
        super().__init__(madlib, model, name)
        self._role = role

    def role(self):
        """
        Returns the role of the role map.
        """
        return self._role

    def value_player(self):
        """
        Returns the value player of the role map.
        """
        return self._role.player()

    @abstractmethod
    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this role map.
        """
        raise NotImplementedError("Subclasses must implement this method")

    def table(self) -> BindableTable:
        """
        Returns the table associated with the role map.
        """
        return self.column().table

    def __repr__(self):
        return f'RoleMap({self._name})'


class InlineValueMap:
    """
    A non-QB relationship class that represents an inline value map.
    """
    def __init__(self, model, binding: Binding, role):
        self._binding = binding
        self._role = role
        self._value_player = role.player()
        self._initialize()

    def _initialize(self):
        role = self._role
        if not role.player()._is_primitive():
            raise TypeError(f'Cannot construct a value map for {role}, concept is not a value type')

        table_name = self._binding.column.table.physical_name()
        concept = role.player()
        _, self._name = _get_map_madlib_and_name(table_name, concept)
        self._has_transforms = bool(self._binding.transform_with)

    def _construct_body_atoms(self, filter_atoms):
        result_type = self.value_player()
        if self._has_transforms:
            transform_result = self._generate_transform_body()
            return result_type(transform_result), [
                where(
                    self._binding.column,  # needed when transform-to-a-literal is used to preserve row_id
                    filter_atoms
                )]
        else:
            return result_type(self._binding.column), [
                filter_atoms,
            ]

    def _generate_transform_body(self):
        transformations = self._binding.transform_with
        if transformations:
            transformations = (transformations,) if not isinstance(transformations, tuple) else transformations
            return self._chain_transformations(self._binding.column, transformations)
        raise ValueError("Incorrect state, cannot apply binding transformations with empty value converter(s)")

    def _chain_transformations(self, col_val_ref, transformations):
        result = col_val_ref
        for transformer in transformations:
            result = self._apply_transformations(result, transformer)
        return result

    @staticmethod
    def _apply_transformations(input_var, transformer):
        if isinstance(transformer, Callable):
            return transformer(input_var)
        else:
            raise TypeError(f'Expected a Relationship or Callable, got {type(transformer)}')

    def value_player(self):
        """
        Returns the value player of the inline value map.
        """
        return self._value_player

    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this inline value map.
        """
        return self._binding.column

    def table(self) -> BindableTable:
        """
        Returns the table associated with the inline value map.
        """
        return self.column().table

    def binding(self) -> Binding:
        """
        Returns the binding of the value map.
        """
        return self._binding

    def formula(self):
        """
        Constructs the body formula for the value map.
        This is used to generate the head and the body of the relation.
        """
        filter_atoms = where(self._binding.filter_by)
        return self._construct_body_atoms(filter_atoms)

    def __repr__(self):
        return f'@inline ValueMap({self._name})'


class ValueMap(RoleMap):
    """
    A class representing a value map relation.
    """

    def __init__(self, model, binding: Binding, role):
        # =
        # Skip QB relationship initialization if *inline* is True, as __call__ will return a QB expression
        # =
        madlib, name = self._handle_params(binding, role)
        super().__init__(madlib, model, name, role)
        self._binding = binding
        self._inline_value_map = InlineValueMap(model, binding, role)
        self._generate_body()

    @staticmethod
    def _handle_params(binding, role):
        if not role.player()._is_primitive():
            raise TypeError(f'Cannot construct a value map for {role}, concept is not a value type')

        table_name = binding.column.table.physical_name()
        concept = role.player()
        return _get_map_madlib_and_name(table_name, concept)

    def _generate_body(self):
        source_ref = self.table()
        val_ref, body_atoms = self._inline_value_map.formula()

        define(self(source_ref, val_ref)).where(*body_atoms)

    def value_player(self):
        return self._inline_value_map.value_player()

    def column(self) -> BindableAttribute:
        return self._inline_value_map.column()

    def binding(self) -> Binding:
        return self._inline_value_map.binding()

    def formula(self):
        source_ref = self.table()
        rez_val_ref = self.value_player().ref()
        return rez_val_ref, [self(source_ref, rez_val_ref)]

    def __repr__(self):
        return f'ValueMap({self._name})'


class MaterializedEntityMap(AbstractMap):
    """
    A class representing a materialized entity map relation.
    This is used to materialize an entity map from a set of role maps.
    """

    def __init__(self, model, inline_entity_map: 'AbstractInlineEntityMap'):
        super().__init__(inline_entity_map.madlib(), model, inline_entity_map.short_name())
        self._inline_entity_map = inline_entity_map
        self._generate_body()

    def _generate_body(self):
        # materialize the population of the entity map as necessary
        self._inline_entity_map.materialize_population()

        source_ref = self.table()
        rez_val_ref, body_formula = self._inline_entity_map.formula()
        where(*body_formula).define(self(source_ref, rez_val_ref))

    def value_player(self):
        """
        Returns the result value player of the materialized entity map.
        This is the same as the value player of the inline entity map.
        """
        return self._inline_entity_map.value_player()

    def column(self) -> BindableAttribute:
        return self._inline_entity_map.column()

    def table(self) -> BindableTable:
        """
        Returns the table associated with the materialized entity map.
        """
        return self._inline_entity_map.table()

    def __repr__(self):
        # replace @inline with @materialized
        return self._inline_entity_map.__repr__().replace('@inline', '@materialized')


class AbstractInlineEntityMap(ABC):
    """
    A non-QB relationship class that represents an inline entity map.
    This is used to construct entities from a set of role maps.
    """

    def __init__(self, model, madlib, name):
        self._model = model
        self._madlib = madlib
        self._name = name

    def __call__(self, *args, **kwargs):
        if bool(args):
            raise ValueError(f'Expected no arguments to a Value Map, got {len(args)}: {args}')

        return self._construct_body_formula()

    def madlib(self):
        """
        Returns the madlib of the inline entity map.
        """
        return self._madlib

    def short_name(self):
        """
        Returns the name of the inline entity map.
        """
        return self._name

    @abstractmethod
    def value_player(self):
        """
        Returns the value player of the role map.
        """
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this role map.
        """
        raise NotImplementedError("Subclasses must implement this method")

    def table(self) -> BindableTable:
        """
        Returns the table associated with the role map.
        """
        return self.column().table

    def formula(self):
        return self._construct_body_formula()

    @abstractmethod
    def _construct_body_formula(self) -> tuple[Union[ConceptConstruct, ConceptFilter, Ref, RelationshipRef], list[typing.Any]]:
        """
        Constructs the body formula for the inline entity map.
        This is used to generate the body of the relation.
        """
        pass

    def materialize(self):
        return MaterializedEntityMap(self._model, self)

    @abstractmethod
    def materialize_population(self):
        """
        Materializes the population of the entity map, using the body of the entity map.
        This has to check the conditions and only update the population if those are met.
        """
        pass


class SimpleConstructorEntityMap(AbstractInlineEntityMap):
    """
    A class representing an entity map relation.
    """

    def __init__(self, model, binding: Binding, role: Role, identifier_to_role_map: OrderedDict[Unique, 'RoleMap']):
        madlib, name = self._handle_params(binding, role)
        super().__init__(model, madlib, name)
        self._binding = binding
        self._role = role
        self._identifier_to_role_map = identifier_to_role_map
        self._reference_role_map = list(identifier_to_role_map.items())[-1][1]  # last role map is the reference one

    @staticmethod
    def _handle_params(binding, role):
        if role.player()._is_primitive():
            raise TypeError(f'Cannot construct an entity map for {role}, concept is not an entity type')

        table_name = binding.column.table.physical_name()
        concept = role.player()
        return _get_map_madlib_and_name(table_name, concept)

    def _construct_body_formula(self):
        concept = self._role.player()
        value_ref, ref_body_atoms = self._reference_role_map.formula()
        entity_ref = reference_entity(concept, value_ref, check_population=False)
        body_atoms = [
            *ref_body_atoms
        ]
        return entity_ref, body_atoms

    def _concept_population_atom(self, entity):
        """
        Generates the *optional* population atom for the entity map.
        Only used for bindings that have a FK.
        """
        if self._should_reference():
            return where(
                self._role.player()(entity)
            )
        else:
            return where()

    def _should_reference(self) -> bool:
        return self._reference_role_map.column().references_column is not None

    def value_player(self):
        """
        Returns the value player of the entity map.
        """
        return self._role.player()

    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this entity map.
        """
        return self._binding.column

    def materialize_population(self):
        concept = self._role.player()
        if not self._should_reference():
            # then create entities
            role_maps = self._identifier_to_role_map.values()
            map_value_refs, subformulas = [], []
            for role_map in role_maps:
                rez_val, subformula = role_map.formula()
                map_value_refs.append(rez_val)
                subformulas.extend(subformula)
            where(
                *subformulas
            ).define(construct_entity(concept, *map_value_refs))

    def __repr__(self):
        return f'@inline CtorEntityMap({self._name})'


class ReferentEntityMap(AbstractInlineEntityMap):
    """
    A class representing a referent entity map relation.
    """

    def __init__(self, model, binding: Binding, role: Role, constructing_role_map: 'Map'):
        madlib, name = self._handle_params(binding, role)
        super().__init__(model, madlib, name)
        self._binding = binding
        self._role = role
        self._constructing_role_map = constructing_role_map

    @staticmethod
    def _handle_params(binding, role):
        if role.player()._is_primitive():
            raise TypeError(f'Cannot construct an entity map for {role}, concept is not an entity type')

        table_name = binding.column.table.physical_name()
        concept = role.player()
        return _get_map_madlib_and_name(f'{table_name}_ref', concept)

    def _construct_body_formula(self):
        concept = self._role.player()
        orig_entity_ref, subformula_atoms = self._constructing_role_map.formula()

        entity_ref = reference_entity(concept, orig_entity_ref)
        return entity_ref, [
            *subformula_atoms,
        ]

    def value_player(self):
        return self._role.player()

    def binding(self) -> Binding:
        """
        Returns the binding of the entity map.
        """
        return self._binding

    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this entity map.
        """
        return self._binding.column

    def materialize_population(self):
        """
        Referent entity maps do not need to materialize population, as they always reference existing entities.
        """
        pass

    def __repr__(self):
        return f'@inline ReferentEntityMap({self._name})'


class CompositeEntityMap(AbstractInlineEntityMap):
    """
    A class representing a composite entity map relation.

    Takes a sequence of EntityMaps and constructs a composite entity type.
    """

    def __init__(self, model, entity_concept: qb.Concept, *role_maps: 'Map'):
        madlib, name = self._handle_params(entity_concept, *role_maps)
        super().__init__(model, madlib, name)
        self._entity_concept = entity_concept
        self._role_maps = list(role_maps)

    def _handle_params(self, entity_concept: qb.Concept, *role_maps: 'Map'):
        if entity_concept._is_primitive():
            raise TypeError(f'Cannot construct a composite entity map for {entity_concept},'
                            f' concept is not an entity type')
        if len(role_maps) < 2:
            raise ValueError('CompositeEntityMap requires at least two EntityMaps')

        role_map = role_maps[0]
        table = role_map.table()
        self._table = table
        return _get_map_madlib_and_name(table.physical_name(), entity_concept)

    def _construct_body_formula(self):
        var_refs, body_atoms = self._body_formula()

        # the rule to populate the entity map
        entity_ref = reference_entity(self._entity_concept, *var_refs, check_population=False)
        return entity_ref, body_atoms

    @staticmethod
    def _should_reference() -> bool:
        """
        Determines if the composite entity map should reference a column.
        Currently not supported fully.
        """
        return False

    def _body_formula(self):
        var_refs, subformulas = zip(*(role_map.formula() for role_map in self._role_maps))
        return var_refs, [
            *(atom for subformula in subformulas for atom in subformula)
        ]

    def value_player(self):
        """
        Returns the value player of the composite entity map.
        """
        return self._entity_concept

    def table(self) -> BindableTable:
        """
        Returns the table associated with the composite entity map.
        """
        return self._table

    def column(self) -> BindableAttribute:
        raise ValueError('CompositeEntityMap does not index a single column')

    def materialize_population(self):
        var_refs, body_atoms = self._body_formula()
        if not self._should_reference():
            # construct entities
            entity_ref = construct_entity(self._entity_concept, *var_refs)
            where(
                *body_atoms,
            ).define(entity_ref)

    def __repr__(self):
        return f'@inline CompositeEntityMap({self._name})'


AbstractEntityMap = Union['SimpleConstructorEntityMap', 'ReferentEntityMap', 'CompositeEntityMap', 'UnionEntityMap']
Map = Union['ValueMap', 'InlineValueMap', 'UnionEntityMap', 'AbstractInlineEntityMap', 'MaterializedEntityMap']

class EntitySubtypeMap(AbstractInlineEntityMap):
    """
    A class representing an entity subtype map relation.
    """

    def __init__(self, model, binding: Union[IdentifierConceptBinding, SubtypeConceptBinding, ReferentConceptBinding],
                 ctor_entity_map: Union[AbstractInlineEntityMap, MaterializedEntityMap]):
        # type of the entity subtype map is the parent type coming from the ctor_entity_map
        madlib, name = self._handle_params(binding, ctor_entity_map.value_player())
        super().__init__(model, madlib, name)
        self._binding = binding
        self._ctor_entity_map = ctor_entity_map

    @staticmethod
    def _handle_params(binding: Union[IdentifierConceptBinding, SubtypeConceptBinding, ReferentConceptBinding],
                       result_concept: qb.Concept):
        table = binding.column.table
        entity_concept = binding.entity_type
        # check that the entity concept is a subtype of the result concept
        assert entity_concept._isa(result_concept), f"Entity concept {entity_concept} must be a subtype of result concept {result_concept}"
        name = f'{table.physical_name()}_row_to_{entity_concept}'
        madlib = f'{name} {{row_id:RowId}} {{val:{result_concept}}}'
        return madlib, name

    def _construct_body_formula(self):
        parent_type_ref, body_atoms = self._generate_body_atoms()

        # note: have to return the parent type here, as otherwise it gets a join on the subtype population
        #       but doesn't cast to the type of the subtype...
        return parent_type_ref, body_atoms

    def _generate_body_atoms(self):
        filtering_atom = where(self._binding.filter_by)
        parent_type_ref, subformula_atoms = self._ctor_entity_map.formula()
        return parent_type_ref, [
            *subformula_atoms,
            filtering_atom
        ]

    def binding(self):
        """
        Returns the binding of the entity subtype map.
        """
        return self._binding

    def value_player(self):
        """
        Returns the subtype of the entity subtype map.
        """
        return self._binding.entity_type

    def column(self) -> BindableAttribute:
        """
        Returns the bindable column associated with this entity subtype map.
        """
        return self._binding.column

    def table(self) -> BindableTable:
        """
        Returns the table associated with the entity subtype map.
        """
        return self._binding.column.table

    def materialize_population(self):
        subtype = self._binding.entity_type
        parent_type_ref, body_atoms = self._generate_body_atoms()
        # TODO instead of Person.new maybe try Employee.new to construct these?
        # derive subtype population
        where(*body_atoms).define(subtype(parent_type_ref))

    def __repr__(self):
        return f'@inline EntitySubtypeMap({self._name})'


class UnionEntityMap(AbstractMap):
    """
    A class representing a union entity map relation.
    """

    def __init__(self, model, entity_concept: qb.Concept, *entity_maps: AbstractEntityMap, generate_population: bool=False):
        madlib, name = self._handle_params(entity_concept, *entity_maps)
        super().__init__(madlib, model, name)
        self._entity_type = entity_concept
        self._entity_maps = OrderedSet().update(entity_maps)
        self._generate_population = generate_population
        self._generate_body()

    def _handle_params(self, entity_concept: qb.Concept, *entity_maps: AbstractEntityMap):
        if len(entity_maps) == 0:
            raise ValueError('UnionEntityMap requires at least one EntityMap')
        # pick an arbitrary entity map to get the table, as they must all have the same
        table = entity_maps[0].table()
        self._table = table
        return _get_map_madlib_and_name(table.physical_name(), entity_concept)

    def _generate_body(self):
        for entity_map in self._entity_maps:
            self._generate_body_rule(entity_map)

    def _generate_body_rule(self, entity_map: AbstractEntityMap):
        source_ref = self.table()
        rez_val_ref, subformula_atoms = entity_map.formula()

        # derive union entity map
        where(
            *subformula_atoms
        ).define(self(source_ref, rez_val_ref))

    def value_player(self):
        """
        Returns the type of the entity map.
        """
        return self._entity_type

    def table(self):
        """
        Returns the table associated with the entity map.
        """
        return self._table

    def update(self, entity_map: AbstractEntityMap):
        """
        Updates the union entity map with a new entity map.
        """
        if entity_map in self._entity_maps:
            return
        self._entity_maps.add(entity_map)
        self._generate_body_rule(entity_map)

    def materialize_population(self):
        if self._generate_population:
            # derive type population
            for entity_map in self._entity_maps:
                rez_val_ref, subformula_atoms = entity_map.formula()
                where(
                    *subformula_atoms
                ).define(self._entity_type(rez_val_ref))

    def __repr__(self):
        return f'UnionEntityMap({self._name})'



# Copyright 2024 RelationalAI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# A Schema is a disjunctions of SchemaFragments, each of which is a conjunction
# of predicate and propositional literals.
#
from abc import abstractmethod
from typing import Literal

from v0.relationalai.early_access.dsl.schemas.builder import SchemaBuilder
from v0.relationalai.early_access.dsl.schemas.contexts import SchemaContext, DASHES, ScalarComponent, RelationalComponent
from v0.relationalai.early_access.dsl.schemas.fragments import SchemaFragment

from sys import stderr

class AbstractSchema:

    def __init__(self): pass


class FalseSchema(AbstractSchema):

    def __init__(self):
        super().__init__()

    def display(self): return "False"

    def pprint(self): return "FALSE"


# Schema
class Schema(AbstractSchema, SchemaBuilder):

    # Each Schema object is a ContextManager
    def __enter__(self):
        SchemaContext.open(self)  # open a new context for this Schema
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        SchemaContext.close()
        self.register_attrs()

    # Construct instances of this schema with an optional set of positional
    # arguments and an optional set of keyword arguments.
    #
    # Special keyword arguments -- DASHES and PREFIX -- decorate the Schema that is
    # being constructed. Other keyword arguments denote the renaming of a component
    # using what would otherwise have been the name of the component in the Schema
    # that is being constructed. CompNames inherit these decorations and renamings
    # from context when a Set is ultimately declared using some name. We implement
    # this inheritance using renaming-, dash-, and prefix- context stacks.
    #
    # Also, each Schema instance will contain an instance variable for each of its
    # components so that contexts that use the instane may refer to the components
    # by name. To make these references safe under schema inclusion, we name the
    # variables using the context-free name of the component, which is the name
    # of the component when the Schema is instantiated at the top level (i.e.,
    # not in the context of some containing or including Schema).
    #
    def __init__(self, *args):
        super().__init__()

        self.disjuncts = []

        # Sometimes a Schema can be instantiated with parameters that are not
        # other Schemas, such as happens when we create a disjunction of the
        # evaluation or reduction of schemas that might reduce to 'None'.
        #
        (parts, params) = partition_schema_args(args)

        # Any parts should be merged in because when a Schema is instantiated with
        # other Schemas as parts, they are to be conjoined by schema inclusion.
        #
        for p in parts:
            self.merge(p)

        if len(params) > 0:
            for p in params:
                if p is not None:
                    print(f"[INFO]: Found non-Schema parameter {p} in instaniation of schema {self}", file=stderr)

        self.register_attrs()

    # Allows subclasses of Schema to generate fresh instances of the subclass
    # so that we can declare more methods here in this abstract superclass
    #
    @abstractmethod
    def _generate(self) -> 'Schema':
        pass

    # Let self be an operation schema and pre an undecorated state schema. This operation
    # returns a schema that is self reduced by constraints that are entailed by what remains
    # in self if we assume pre.
    #
    def assume_precondition(self, pre):
        new_schema: Schema = self._generate()
        for t in self.disjuncts:
            for p in pre.disjuncts:
                new_schema.disjuncts.append(t.assume_precondition(p))
        return new_schema

    def build_relation_variable(self, args, kwargs):

        dashes = kwargs[DASHES] if DASHES in kwargs else 0

        tp = args[0]
        name = args[1]

        if self.disjunctive():
            raise Exception("Cannot add a component directly to a disjunctive schema")

        return self.declare_relation(name, tp, dashes)


    def build_scalar_variable(self, args, kwargs):

        dashes = kwargs[DASHES] if DASHES in kwargs else 0

        tp = args[0]
        name = args[1]
        if self.disjunctive():
            raise Exception("Cannot add a component directly to a disjunctive schema")

        return self.declare(name, tp, dashes)


    # Causes this Schema to declare a new ScalarComponent of the given name, type,
    # and decorartion (number of dashes)
    #
    def declare(self, name, type, dashes=0):
        decl = ScalarComponent(name, type, dashes)

        self.add_component(decl)

        v = decl.basevar()
        setattr(self, v.name(), v)

        return decl

    # Causes this Schema to declare a new RelationalComponent of the given name,
    # signature, and decorartion (number of dashes)
    #
    def declare_relation(self, name, signature, dashes=0):
        decl = RelationalComponent(name, signature, dashes)

        self.add_relational_component(decl)
        v = decl.basevar()
        setattr(self, v.name(), v)

        return decl

    def decorate(self):
        new_schema = self._generate()
        for c in self.disjuncts:
            new_schema.disjuncts.append(c.decorate())
        new_schema.register_attrs()
        return new_schema

    # Given a tuple of relational vars, return a new Schema that is a duplicate
    # of self with RelationIsEmpty constraints on all v in vars
    #
    def is_empty(self, *vars):
        new_schema = self._generate()
        for c in self.disjuncts:
            new_schema.disjuncts.append(c.is_empty(*vars))
        new_schema.register_attrs()
        return new_schema

    def register_attrs(self):
        attrs = {}
        for d in self.disjuncts:
            for c in d._scalars.values():
                v = c.basevar()
                attrs[v.name()] = v
            for c in d._relations.values():
                v = c.basevar()
                attrs[v.name()] = v

        for v in attrs.values():
            setattr(self, v.name(), v)

    def constructor(self, *args):
        pass

    def disjunctive(self) -> bool:
        return False

    @abstractmethod
    def merge(self, other) -> 'Schema':
        pass

    # Let self be some operation Schema. For each dashed relation R' in self, if there
    # exists at least one AtomicConstraint involving R' in the constraints of self, we can
    # generate a (potentially recursive) Rel rule that populates R by conjoining the
    # constraints that refer either directly or indirectly to R.
    #
    def genrel(self):
        operation = []

        reduced = self.reduce()
        for s in reduced.disjuncts:
            operation.append(s.refine().pprint())
        return "\n\n".join(operation)

    # Creates a new schema by hiding variables named in args under an existential
    # quantifier. This is just a wrapper to conceal that allows users to directly
    # reference Schema instance variables when declaring Schema classes
    #
    def hide(self, *args):
        if len(args) == 0:
            raise Exception("Must name variables to hide from a schema")

        return self.conceal([var.display() for var in args])

    # Add SchemaComponent d to the declaration part of this Schema
    #
    def add_component(self, d):
        if len(self.disjuncts) == 0:
            self.disjuncts.append(SchemaFragment())

        # By default, we assume that we will need to cache component d.
        for disj in self.disjuncts:
            disj.add_component(d)

    # Add SchemaComponent d to the declaration part of this Schema
    #
    def add_relational_component(self, d):
        if len(self.disjuncts) == 0:
            self.disjuncts.append(SchemaFragment())

        # By default, we assume that we will need to cache component d.
        for disj in self.disjuncts:
            disj.add_relational_component(d)

    def add_scalar_constraint(self, c):
        if len(self.disjuncts) == 0:
            self.disjuncts.append(SchemaFragment())

        for disj in self.disjuncts:
            disj.add_scalar_constraint(c)

    def add_atomic_constraint(self, c):
        if len(self.disjuncts) == 0:
            self.disjuncts.append(SchemaFragment())

        for disj in self.disjuncts:
            disj.add_atomic_constraint(c)

    def add_predicate_constraint(self, c):
        if len(self.disjuncts) == 0:
            self.disjuncts.append(SchemaFragment())

        for disj in self.disjuncts:
            disj.add_predicate_constraint(c)

    # Returns the schema formed by concealing (hiding) variables in the list cvars
    # under an existential quantifier.
    #
    # Note: Clients typically call this method indirectly from the hide method.
    #
    def conceal(self, cvars):
        new_schema = self._generate()
        for disj in self.disjuncts:
            new_schema.disjuncts.append(disj.conceal(cvars))
        return new_schema

    # Partially evaluate this Schema by substituting statically-computable values for
    # its schema variables, simplifying where possible.
    #
    def evaluate(self, bindings):
        new_schema = self._generate()
        for disj in self.disjuncts:
            d = disj.evaluate(bindings)
            if d is not None:
                new_schema.disjuncts.append(d)

        # Check if evaluation violates some schema constraint
        if len(self.disjuncts) > 0 and len(new_schema.disjuncts) == 0:
            return FalseSchema()

        return new_schema

    # Emit this schema to a textual output using the Z display style
    #
    def pprint(self):
        declStrings = []

        declStrings.append("( ")
        disjunctStrings = []
        for disj in self.disjuncts:
            disjunctStrings.append(disj.pprint())
        declStrings.append("\n OR \n".join(disjunctStrings))
        declStrings.append(" )")
        return "\n".join(declStrings)

    def prefix(self, pref):
        new_schema = self._generate()
        for c in self.disjuncts:
            new_schema.disjuncts.append(c.prefix(pref))
        new_schema.register_attrs()
        return new_schema

    # Simplify this Schema by checking the satisfiability of its propositional
    # constraints.
    #
    def reduce(self):
        new_schema = self._generate()
        for disj in self.disjuncts:
            s = disj.reduce()
            if s is not None:
                new_schema.disjuncts.append(s)
        return new_schema

    # Simplify this Schema by projecting out decls that are equated with
    # other decls via predicate constraints. These commonly arise when
    # decorating Schemas using the Xi operator. Note that when reducing
    # we try to remove the SchemaComponent with the larger number of dashes
    # to simplify Rel rule generation.
    #
    def reduce_by_xi(self):
        new_schema = self._generate()
        for disj in self.disjuncts:
            new_schema.disjuncts.append(disj.reduce_by_xi())
        return new_schema

    # Create a new Schema by renaming this Schema's decls.
    # Each arg must be a dictionary that maps component name
    # to some compatible component.
    #
    def rename(self, *args):
        renaming = {}
        for a in args:
            renaming = renaming | a

        new_schema = self._generate()
        for disj in self.disjuncts:
            new_schema.disjuncts.append(disj.rename(renaming))
        new_schema.register_attrs()
        return new_schema

    def sync(self, *args):
        return self.rename(*args)

    # Generates the rule to populate a Rel view that materializes this
    # Schema under the given view name.
    #
    # The rule will be annotated @inline if we cannot statically
    # determine that the Schema is grounded.
    #
    def view(self, viewName):
        s = self.reduce_by_xi().reduce()
        rules = []

        for disj in s.disjuncts:
            rules.append(disj.view(viewName))

        return "\n".join(rules)


class ConjunctiveSchema(Schema):

    def __init__(self, *args):
        Schema.__init__(self, *args)

    def _generate(self): return ConjunctiveSchema()

    # Looks to see if this Schema declares a RelationComponent by this
    # name
    def has_relation_component(self, name) -> bool:
        for d in self.disjuncts:
            if name in d._relations:
                return True
        return False

    # Assuming this Schema already contains a RelationComponent by this
    # name, retrieve and return it to the caller. We use this to support
    # dynamically adding components to a schema when we see references
    # to relations that are defined in an ontology.
    #
    def retrieve_relation_component(self, name):
        for d in self.disjuncts:
            if name in d._relations:
                return d._relations[name]
        return None

    # Merge self with other, mutating self.
    #
    def merge(self, other):
        oldSelfDisjuncts = self.disjuncts
        self.disjuncts = []
        if len(oldSelfDisjuncts) >= 1:
            for d in oldSelfDisjuncts:
                for d2 in other.disjuncts:
                    d3 = d.duplicate()
                    for c in d2._scalars:
                        d3.add_component(d2._scalars[c])
                    for c in d2._relations:
                        d3.add_relational_component(d2._relations[c])
                    for c in d2._sconstraints:
                        d3.add_scalar_constraint(d2._sconstraints[c])
                    for c in d2._atoms:
                        d3.add_atomic_constraint(d2._atoms[c])
                    for c in d2._universals:
                        d3.add_predicate_constraint(d2._universals[c])
                    self.disjuncts.append(d3)
        else:
            self.disjuncts = [d.duplicate() for d in other.disjuncts]
        return self


class DisjunctiveSchema(Schema):

    def __init__(self, *args):
        Schema.__init__(self, *args)

    def _generate(self): return DisjunctiveSchema()

    def disjunctive(self) -> Literal[True]: return True

    def merge(self, other):
        for d in other.disjuncts:
            self.disjuncts.append(d)
        self.register_attrs()
        return self


# When class Schema is instantiated, other Schema instances (parts) can
# be specified, as can arbitrary arguments that are used to parameterize
# the Schema.
#
def partition_schema_args(args):
    parts = []
    params = []
    for a in args:
        if isinstance(a, Schema):
            parts.append(a)
        else:
            params.append(a)
    return (parts, params)



from __future__ import annotations
from collections import deque
from dataclasses import dataclass, field
from sqlglot import exp, select

from ....util.structures import OrderedSet
from ...metamodel.metamodel import Aggregate, Logical, Lookup, Model, Node, Relation, Task, Update, Var, Value
from ...metamodel.pprint import pprint as mmpp
from ...metamodel.rewriter import Rewriter, Walker
from ...metamodel.metamodel_analyzer import Analyzer as MMAnalyzer

from rich import print as rp

#------------------------------------------------------
# Simple flatten
#------------------------------------------------------

class SimpleFlattener(Walker):
    def flatten(self, node: Node) -> list[Task]:
        self.rules: list[Task] = []
        self.stack: list[Task] = []
        self(node)
        return self.rules

    def enter_logical(self, logical: Logical):
        self.stack.append(logical)

    def logical(self, logical: Logical):
        self.stack.pop()

    def update(self, update: Update):
        # build a rule for just this update
        body = OrderedSet()
        for frame in self.stack:
            if isinstance(frame, Logical):
                for task in frame.body:
                    if not isinstance(task, (Logical, Update)):
                        body.add(task)
        body.add(update)
        self.rules.append(Logical(body=tuple(body)))

#------------------------------------------------------
# SourceWalker
#------------------------------------------------------

class SourceWalker(Walker):
    def capture_sources(self, node: Task, analysis:SQLAnalysis):
        self.stack = [OrderedSet()]
        self.root = node
        self.analysis = analysis
        self(node)

    def enter_logical(self, logical: Logical):
        self.stack.append(OrderedSet())

    def logical(self, logical: Logical):
        self.stack.pop()

    def lookup(self, lookup: Lookup):
        self.analysis.get(lookup.relation)
        self.stack[-1].add(lookup.relation)

    def update(self, update: Update):
        info = self.analysis.get(update.relation)
        info.sources.add(self.root)
        for frame in self.stack:
            for dep in frame:
                info.edges.add(dep)
                self.analysis.get(dep)

#------------------------------------------------------
# RecursionWalker
#------------------------------------------------------

class RecursionWalker(Walker):
    def is_simple(self, node: Task, analysis: SQLAnalysis) -> bool:
        self._recursive_count = 0
        self._has_agg = False
        self.analysis = analysis
        self(node)
        if not self._has_agg and self._recursive_count <= 1:
            return True
        return False

    def aggregate(self, aggregate: Aggregate):
        self._has_agg = True

    def lookup(self, lookup: Lookup):
        if self.analysis.get(lookup.relation).recursive:
            self._recursive_count += 1

#------------------------------------------------------
# VarMapper
#------------------------------------------------------

class VarMapper(Rewriter):
    def map_vars(self, mapping: dict[Var, Value], node: Node):
        self.mapping = mapping
        return self(node)

    def var(self, var: Var):
        if var in self.mapping:
            return self.mapping[var]
        return var

#------------------------------------------------------
# Analysis
#------------------------------------------------------

mapper = VarMapper()

@dataclass
class RelationInfo:
    relation: Relation
    sources: OrderedSet[Task] = field(default_factory=OrderedSet[Task])
    edges: OrderedSet[Relation] = field(default_factory=OrderedSet[Relation])

    head_vars: list[Var] = field(default_factory=list)
    body: OrderedSet[Task] = field(default_factory=OrderedSet[Task])

    recursive: bool = field(default=False)

    def map_lookup(self, lookup: Lookup):
        mapping = {self.head_vars[i]: lookup.args[i] for i in range(len(self.head_vars))}
        mapped = []
        for task in self.body:
            mapped.append(mapper.map_vars(mapping, task))
        return mapped

@dataclass
class SQLAnalysis:
    relations: dict[Relation, RelationInfo] = field(default_factory=dict)
    source_walker = SourceWalker()

    def get(self, relation: Relation) -> RelationInfo:
        if relation not in self.relations:
            self.relations[relation] = RelationInfo(relation=relation)
        return self.relations[relation]

    def analyze(self, task:Task):
        self.source_walker.capture_sources(task, self)

    def construct_sccs(self):
        sccs = find_sccs(self.relations)
        self.sccs = sccs
        for info in self.relations.values():
            info.recursive = False
        for scc in sccs:
            for relation in scc.relations:
                self.get(relation).recursive = scc.recursive

        scc_order = scc_toposort(sccs, self.relations)
        self.scc_order = scc_order

#------------------------------------------------------
# SCCs
#------------------------------------------------------

@dataclass
class SCC:
    relations: list[Relation]
    recursive: bool

    def strategy(self, analysis:SQLAnalysis) -> str:
        if not self.recursive:
            return "NORMAL"
        relations = self.relations
        if len(relations) == 1:
            rel = relations[0]
            sources = analysis.get(rel).sources
            if all(RecursionWalker().is_simple(source, analysis) for source in sources):
                return "CTE"
        return "SEMI_NAIVE"

def find_sccs(relations: dict[Relation, RelationInfo]) -> list[SCC]:
    # find sccs using Tarjan's algorithm
    index = 0
    stack: list[Relation] = []
    onstack: set[Relation] = set()
    ix: dict[Relation, int] = {}
    low: dict[Relation, int] = {}
    results: list[SCC] = []

    def strongconnect(v: Relation):
        nonlocal index
        ix[v] = low[v] = index
        index += 1
        stack.append(v)
        onstack.add(v)

        info = relations.get(v)
        edges = info.edges if info else ()
        for w in edges:
            if w not in ix:
                strongconnect(w)
                low[v] = low[v] if low[v] < low[w] else low[w]
            elif w in onstack:
                low[v] = low[v] if low[v] < ix[w] else ix[w]

        if low[v] == ix[v]:
            comp: list[Relation] = []
            while True:
                w = stack.pop()
                onstack.discard(w)
                comp.append(w)
                if w == v:
                    break
            # size > 1 or a singleton with self-loop
            comp_info = relations.get(comp[0])
            comp_edges = comp_info.edges if comp_info else ()
            recursive = (len(comp) > 1) or (comp[0] in comp_edges)
            results.append(SCC(comp, recursive))

    # visit all vertices, including isolated
    nodes = set(relations.keys()) | {w for info in relations.values() for w in info.edges}
    for v in nodes:
        if v not in ix:
            strongconnect(v)

    return results

def scc_toposort(components: list[SCC], relations: dict[Relation, RelationInfo]) -> list[SCC]:
    comp_id = {v: i for i, comp in enumerate(components) for v in comp.relations}
    n = len(components)

    # Build reversed condensed DAG: dep_comp -> rel_comp
    adj: list[set[int]] = [set() for _ in range(n)]
    indeg: list[int] = [0] * n

    for rel, info in relations.items():
        cr = comp_id[rel]
        for dep in info.edges:
            cd = comp_id.get(dep)
            if cd is None:
                continue
            if cr == cd:
                continue
            # edge: dependency SCC -> relation SCC
            if cr not in adj[cd]:
                adj[cd].add(cr)
                indeg[cr] += 1

    # Kahn’s algorithm
    q: deque[int] = deque(i for i, d in enumerate(indeg) if d == 0)
    order: list[SCC] = []
    while q:
        i = q.popleft()
        order.append(components[i])
        for j in adj[i]:
            indeg[j] -= 1
            if indeg[j] == 0:
                q.append(j)

    assert len(order) == n, "SCC DAG unexpectedly cyclic"
    return order

#------------------------------------------------------
# Context
#------------------------------------------------------

class Frame:
    pass

class Context:
    pass

#------------------------------------------------------
# Compiler
#------------------------------------------------------

class SQLCompiler:

    def compile(self, model:Model):
        MMAnalyzer().analyze(model)
        analysis = SQLAnalysis()

        rules = []
        flattener = SimpleFlattener()
        # mmpp(model)

        assert isinstance(model.root, Logical)
        for sub_task in model.root.body:
            flat_rules = flattener.flatten(sub_task)
            rules.extend(flat_rules)
            for rule in flat_rules:
                analysis.analyze(rule)

        analysis.construct_sccs()
        sccs = analysis.sccs

        # print("\nStrongly Connected Components:")
        # for i, scc in enumerate(sccs, start=1):
        #     tag = f" (recursive {scc.strategy(analysis)})" if scc.recursive else ""
        #     print(f"  SCC {i}:{tag}")
        #     for r in scc.relations:
        #         print(f"    - {r}")

        # # ---- Topological order of SCCs (evaluation phases) ----
        # ordered_sccs = analysis.scc_order

        # # TODO: handle relations that didn't end up in any SCC?

        # # walk the ordered components forward collapsing as we go
        # print("\nSCC evaluation order (topologically sorted):")
        # for phase_idx, scc in enumerate(ordered_sccs, start=1):
        #     tag = f" (recursive {scc.strategy(analysis)})" if scc.recursive else ""
        #     rp(f"[yellow bold]  Phase {phase_idx}:{tag}")
        #     rp("[yellow bold]------------------------------------------------------------")
        #     for relation in scc.relations:
        #         info = analysis.get(relation)
        #         rp(f"[cyan bold]  {relation}")
        #         # print("  Head Vars:")
        #         # for hv in info.head_vars:
        #         #     print(f"    - {hv}")
        #         # print("  Body:")
        #         # for task in info.body:
        #         #     print(f"    - {task}")
        #         print("    Sources:")
        #         for source in info.sources:
        #             # print(f"        - {source}")
        #             mmpp(source, indent=2)
        #         print("    Edges:")
        #         for edge in info.edges:
        #             print(f"        - {edge}")
        #     print("\n")





from __future__ import annotations

from collections import defaultdict
from functools import reduce
import re
from typing import Any, Dict, List, Tuple, cast, Optional
import base64

import numpy as np
from pandas import DataFrame
from pandas.api.extensions import ExtensionDtype, ExtensionArray, register_extension_dtype
import pandas as pd
import decimal
from v0.relationalai.tools.constants import Generation
from v0.relationalai import debugging
from .. import metamodel as m

#--------------------------------------------------
# Constants
#--------------------------------------------------

UNIXEPOCH = 62135683200000
MILLISECONDS_PER_DAY = 24 * 60 * 60 * 1000

#--------------------------------------------------
# Custom Int128 dtype
#--------------------------------------------------

@register_extension_dtype
class Int128Dtype(ExtensionDtype):
    name = "Int128"
    type = int

    @ExtensionDtype.kind.getter
    def kind(self):
        return "i"

    @ExtensionDtype.na_value.getter
    def na_value(self):
        return pd.NA

    @classmethod
    def construct_array_type(cls):
        return Int128Array


class Int128Array(ExtensionArray):
    def __init__(self, values):
        # Convert None/pd.NA/np.nan to pd.NA, keep ints as Python int
        self._data = np.array(
            [pd.NA if v is None or v is pd.NA or (isinstance(v, float) and np.isnan(v)) else int(v)
             for v in values],
            dtype=Int128Dtype
        )

    @classmethod
    def _from_sequence(cls, scalars, dtype=None, copy=False):
        return cls(scalars)

    @classmethod
    def _from_factorized(cls, values, original):
        return Int128Array(values)

    @classmethod
    def _concat_same_type(cls, to_concat):
        # to_concat is a sequence of your arrays
        data = np.concatenate([x._data for x in to_concat])
        return cls(data)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._data[item]
        return type(self)(self._data[item])

    def __len__(self):
        return len(self._data)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Int128Array):
            return False
        return self._data == other._data

    @property
    def dtype(self):
        return Int128Dtype()

    @property
    def nbytes(self):
        return self._data.nbytes

    def isna(self):
        return np.array([x is pd.NA for x in self._data])

    def take(self, indexer, *, allow_fill=False, fill_value=None):
        from pandas.api.extensions import take
        if allow_fill and fill_value is None:
            fill_value = self.dtype.na_value
        return type(self)(take(self._data, indexer, fill_value=fill_value, allow_fill=allow_fill))

    def copy(self):
        return type(self)(self._data.copy())

    def __repr__(self):
        return f"Int128Array({list(self._data)})"

    def __add__(self, other):
        if isinstance(other, Int128Array):
            other = other._data
        return Int128Array(
            [pd.NA if a is pd.NA or b is pd.NA else a + b for a, b in zip(self._data, other)]
        )


#--------------------------------------------------
# Result formatting
#--------------------------------------------------

def format_value(value, type):
    if "UInt128" in type:
        return decode_uint128(value)
    elif "FixedDecimal" in type:
        decimal_info = re.search(r"FixedDecimal\{Int(\d+), (\d+)\}", type)
        if decimal_info:
            bits = int(decimal_info.group(1))
            scale = int(decimal_info.group(2))
            if bits == 128:
                return (decimal.Decimal(str((int(value[1]) << 64) + int(value[0]))) if value[1] > 0 else decimal.Decimal(str(value[0]))) / (10 ** scale)
            else:
                return decimal.Decimal(str(value)) / (10 ** scale)
    elif "Int128" in type:
        return (int(value[1]) << 64) + int(value[0]) if value[1] > 0 else value[0]
    elif "Missing" in type:
        return None
    return value

def format_columns(result_frame:DataFrame, types:List[str], generation:Optional[Generation]=Generation.V0):
    for i, col in enumerate(result_frame.columns):
        if "UInt128" in types[i]:
            result_frame[col] = result_frame[col].apply(decode_uint128)
        elif "FixedDecimal" in types[i]:
            decimal_info = re.search(r"FixedDecimal\{Int(\d+), (\d+)\}", types[i])
            if decimal_info:
                bits = int(decimal_info.group(1))
                scale = int(decimal_info.group(2))
                if bits == 128:
                    result_frame[col] = result_frame[col].apply(lambda x: decode_fixed_decimal_128(x, scale))
                else:
                    result_frame[col] = result_frame[col].apply(lambda x: decode_fixed_decimal(x, scale))
        elif "Int128" in types[i]:
            if generation == Generation.V0:
                result_frame[col] = result_frame[col].apply(decode_int128)
            else:
                result_frame[col] = result_frame[col].apply(decode_int128).astype(Int128Dtype())
        elif "Missing" in types[i]:
            result_frame[col] = result_frame[col].apply(lambda _: None)
        elif types[i] == "Dates.DateTime":
            result_frame[col] = pd.to_datetime(result_frame[col] - UNIXEPOCH, unit="ms")
        elif types[i] == "Dates.Date":
            result_frame[col] = pd.to_datetime(result_frame[col] * MILLISECONDS_PER_DAY - UNIXEPOCH, unit="ms")
        elif types[i] == "Int64":
            result_frame[col] = result_frame[col].astype('Int64')
    return result_frame

def sort_data_frame_result(df:DataFrame):
    try:
        df.sort_values(by=[str(c) for c in df.columns], ascending=[True] * len(df.columns), inplace=True)
    except Exception:
        pass
    return df.reset_index(drop=True)

def decode_int128(x):
    high, low = int(x[1]), int(x[0])
    decoded = (high << 64) + low if high != 0 else low
    # For negative numbers, we need to apply two's complement
    if high & (1 << 63): # MSB is sign bit
        decoded = decoded - (1 << 128)
    return decoded

dec128ctx = decimal.Context(prec=38, rounding=decimal.ROUND_HALF_EVEN)
def decode_fixed_decimal_128(x: Any, scale: int) -> Any:
    high, low = int(x[1]), int(x[0])
    value = (high << 64) + low

    # For negative numbers, we need to apply two's complement
    if high & (1 << 63): # MSB is sign bit
        value = value - (1 << 128)

    return decimal.Decimal(value, context=dec128ctx).scaleb(-scale, context=dec128ctx)

def decode_fixed_decimal(x: Any, scale: int) -> Any:
    scale_format = decimal.Decimal(1).scaleb(-scale)
    factor = decimal.Decimal(10) ** scale

    return (decimal.Decimal(str(x)) / factor).quantize(scale_format, rounding=decimal.ROUND_HALF_EVEN)


def decode_uint128(x):
    if isinstance(x, dict):
        keys = sorted(x.keys())
        result = []
        for key in keys:
            value = x[key]
            if isinstance(value, np.ndarray):
                result.extend(value)
            elif isinstance(value, (int, np.integer)):
                result.append(value)
            else:
                raise ValueError(f"Unsupported type for key {key}: {type(value)}")

        return decode_uint128(np.array(result, dtype=np.uint64))

    elif isinstance(x, np.ndarray):
        return base64.b64encode(x.tobytes()).decode()[:-2]

    else:
        raise ValueError(f"Unsupported input type: {type(x)}")

def convert_hash_value(value):
    if isinstance(value, np.ndarray):
        return tuple(value)
    return value

def merge_columns(left:DataFrame, right:DataFrame, on:List[str]):
    if not len(left.columns):
        for col in on:
            left.insert(loc=len(left.columns), column=col, value=np.nan)
        col_count = len(left.columns)
        left.insert(loc=col_count, column=f'new_column{col_count}', value=np.nan)
    if right.empty:
        col_count = len(left.columns)
        left.insert(loc=col_count, column=f'new_column{col_count}', value=np.nan)
        return left
    if not on:
        return pd.merge(left, right, how='cross')
    return pd.merge(left, right, on=on, how='outer')

def format_results(results, task:m.Task|None, result_cols:List[str]|None = None, generation:Optional[Generation]=Generation.V0) -> Tuple[DataFrame, List[Any]]:
    with debugging.span("format_results"):
        data_frame = DataFrame()
        problems = defaultdict(
            lambda: {
                "message": "",
                "path": "",
                "start_line": None,
                "end_line": None,
                "report": "",
                "code": "",
                "severity": "",
                "decl_id": "",
                "name": "",
                "output": "",
                "end_character": None,
                "start_character": None,
                "props": {},
            }
        )

        # Check if there are any results to process
        if len(results.results):
            ret_cols = result_cols or (task.return_cols() if task else [])
            has_cols:List[DataFrame] = [DataFrame() for _ in range(0, len(ret_cols))]
            key_len = 0
            for result in results.results:
                relation_id = result["relationId"]
                result_frame = result["table"].to_pandas()

                types = [
                    t
                    for t in result["relationId"].split("/")
                    if t != "" and not t.startswith(":")
                ]

                # Process diagnostics
                if "/:rel/:catalog/:diagnostic/" in relation_id:
                    # Handle different types of diagnostics based on relation_id
                    if "/:message/" in relation_id:
                        for _, row in result_frame.iterrows():
                            if len(row) > 1:
                                problems[row.iloc[0]]["message"] = row.iloc[1]
                    elif "/:report/" in relation_id:
                        for _, row in result_frame.iterrows():
                            if len(row) > 1:
                                problems[row.iloc[0]]["report"] = row.iloc[1]
                    elif "/:start/:line" in relation_id:
                        for _, row in result_frame.iterrows():
                            if len(row) > 2:
                                problems[row.iloc[0]]["start_line"] = row.iloc[2]
                    elif "/:end/:line" in relation_id:
                        for _, row in result_frame.iterrows():
                            if len(row) > 2:
                                problems[row.iloc[0]]["end_line"] = row.iloc[2]
                    elif "/:model" in relation_id:
                        for _, row in result_frame.iterrows():
                            if len(row) > 1:
                                problems[row.iloc[0]]["path"] = row.iloc[1]
                    elif "/:severity" in relation_id:
                        for _, row in result_frame.iterrows():
                            if len(row) > 1:
                                problems[row.iloc[0]]["severity"] = row.iloc[1]
                    elif "/:code" in relation_id:
                        for _, row in result_frame.iterrows():
                            if len(row) > 1:
                                problems[row.iloc[0]]["code"] = row.iloc[1]

                # Process integrity constraint violations
                elif "/:rel/:catalog/:ic_violation" in relation_id:
                    if "/:decl_id" in relation_id:
                        for _, row in result_frame.iterrows():
                            if len(row) > 1:
                                problems[convert_hash_value(row.iloc[0])]["decl_id"] = (
                                    row.iloc[1]
                                )
                                problems[convert_hash_value(row.iloc[0])]["code"] = (
                                    "IC_VIOLATION"
                                )
                    elif "/:model" in relation_id:
                        for _, row in result_frame.iterrows():
                            if len(row) > 1:
                                problems[convert_hash_value(row.iloc[0])]["path"] = (
                                    row.iloc[1]
                                )
                    elif "/:name" in relation_id:
                        # Get the last segment of the relation_id as the name
                        segments = [
                            segment[1:]
                            for segment in relation_id.split("/")[4:]
                            if segment.startswith(":")
                        ]
                        for _, row in result_frame.iterrows():
                            problems[convert_hash_value(row.iloc[0])]["name"] = segments[-1]
                    elif "/:output" in relation_id:
                        for _, row in result_frame.iterrows():
                            if len(row) > 1:
                                problems[convert_hash_value(row.iloc[0])]["message"] = (
                                    row.iloc[1]
                                )
                    elif "/:range/:end/:character" in relation_id:
                        for _, row in result_frame.iterrows():
                            if len(row) > 1:
                                problems[convert_hash_value(row.iloc[0])][
                                    "end_character"
                                ] = row.iloc[1]
                    elif "/:range/:end/:line" in relation_id:
                        for _, row in result_frame.iterrows():
                            if len(row) > 1:
                                problems[convert_hash_value(row.iloc[0])]["end_line"] = (
                                    row.iloc[1]
                                )
                    elif "/:range/:start/:character" in relation_id:
                        for _, row in result_frame.iterrows():
                            if len(row) > 1:
                                problems[convert_hash_value(row.iloc[0])][
                                    "start_character"
                                ] = row.iloc[1]
                    elif "/:range/:start/:line" in relation_id:
                        for _, row in result_frame.iterrows():
                            if len(row) > 1:
                                problems[convert_hash_value(row.iloc[0])]["start_line"] = (
                                    row.iloc[1]
                                )
                    elif "/:report" in relation_id:
                        for _, row in result_frame.iterrows():
                            if len(row) > 1:
                                problems[convert_hash_value(row.iloc[0])]["report"] = (
                                    row.iloc[1]
                                )

                elif "/:pyrel_error" in relation_id:
                    for _, row in result_frame.iterrows():
                        id = convert_hash_value(row.iloc[0])
                        problems[id]["code"] = "PYREL_ERROR"
                        if row.iloc[1] == "message":
                            problems[id]["message"] = row.iloc[2]
                        elif row.iloc[1] == "severity":
                            problems[id]["severity"] = row.iloc[2]
                        else:
                            props = cast(Dict, problems[id]["props"])
                            props[row.iloc[1]] = format_value(row.iloc[2], relation_id.split("/")[-1])

                elif "/:__pyrel_debug_watch" in relation_id:
                    result_frame = format_columns(result_frame, types, generation)
                    from v0.relationalai.experimental.inspect import _print_watch_frame
                    _print_watch_frame(result_frame)

                # Process other results
                else:
                    result_frame = format_columns(result_frame, types, generation)
                    result["table"] = result_frame
                    if ":output/:cols" in result["relationId"]:
                        matched = re.search(r":col([0-9]+)", result["relationId"])
                        assert matched, f"Could not determine column mapping for {result['relationaId']}"
                        col_ix = int(matched.group(1))
                        key_cols = [
                            f"id{i}" for i in range(0, len(result_frame.columns) - 1)
                        ]
                        key_len = len(key_cols)
                        result_frame.columns = [*key_cols, f"v{col_ix}"]
                        if has_cols[col_ix].empty:
                            has_cols[col_ix] = result_frame
                        else:
                            has_cols[col_ix] = pd.concat([has_cols[col_ix], result_frame], ignore_index=True)
                    elif ":output" in result["relationId"]:
                        data_frame = pd.concat(
                            [data_frame, result_frame], ignore_index=True
                        )

            if any(not col.empty for col in has_cols):
                key_cols = [f"id{i}" for i in range(0, key_len)]
                df_wide_reset = reduce(lambda left, right: merge_columns(left, right, key_cols), has_cols)
                data_frame = df_wide_reset.drop(columns=key_cols)

            data_frame = sort_data_frame_result(data_frame)

            if len(ret_cols) and len(data_frame.columns) == len(ret_cols):
                if task is not None:
                    data_frame.columns = task.return_cols()[: len(data_frame.columns)]
                elif result_cols is not None:
                    data_frame.columns = result_cols[: len(data_frame.columns)]

        return (data_frame, list(problems.values()))



from __future__ import annotations
from typing import Optional

from v0.relationalai.semantics.metamodel import ir, factory as f, builtins as bt, types, helpers
from v0.relationalai.semantics.metamodel.compiler import Pass
from v0.relationalai.semantics.metamodel.util import OrderedSet, ordered_set
from v0.relationalai.semantics.metamodel import dependency

class CDC(Pass):
    """
    Pass to process tables brought to Relational AI logical engines by CDC. When CDC occurs,
    wide snowflake tables are shredded into smaller tables. This pass ensures that code that
    reads from the wide relation is changed to read from the smaller tables. Furthermore,
    it attaches the @function annotation to the property lookups, as an optimization.

    Beware that this pass makes assumptions about the names and types of CDC relations and
    columns!

    From:
        Logical
            TPCH.SF1.LINEITEM(l_orderkey, l_partkey, l_suppkey, l_linenumber, l_quantity, l_extendedprice, l_discount, l_tax, l_returnflag, l_linestatus, l_shipdate, l_commitdate, l_receiptdate, l_shipinstruct, l_shipmode, l_comment)
            construct(LineItem, "l_orderkey", l_orderkey, "l_linenumber", l_linenumber, lineitem)
            -> derive LineItem(lineitem)
            -> derive l_orderkey(lineitem, l_orderkey)
            -> derive l_linenumber(lineitem, l_linenumber)
    To:
    Logical
        tpch_sf1_lineitem("L_ORDERKEY", row_id, l_orderkey)
        tpch_sf1_lineitem("L_LINENUMBER", row_id, l_linenumber)
        construct(LineItem, "l_orderkey", l_orderkey, "l_linenumber", l_linenumber, lineitem)
        -> derive LineItem(lineitem)
        -> derive l_orderkey(lineitem, l_orderkey) (@function)
        -> derive l_linenumber(lineitem, l_linenumber) (@function)
    """

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        # create the dependency analysis context
        ctx = CDC.CDCContext(model)

        # rewrite the root
        replacement = self.handle(model.root, ctx)

        # the new root contains the extracted top level logicals and the rewritten root
        if ctx.rewrite_ctx.top_level:
            new_root = ir.Logical(model.root.engine, tuple(), tuple(ctx.rewrite_ctx.top_level + [replacement]))
        else:
            new_root = replacement

        # create the new model, updating relations and root
        return ir.Model(
            model.engines,
            OrderedSet.from_iterable(model.relations).update(ctx.rewrite_ctx.relations).frozen(),
            model.types,
            new_root
        )

    class CDCContext():
        def __init__(self, model: ir.Model):
            self.model = model
            self.rewrite_ctx = helpers.RewriteContext()
            self.cdc_relations = dict()
            self.info: Optional[dependency.DependencyInfo] = None

        def get_info(self):
            """ Lazily compute info as needed. """
            if self.info is None:
                # no need for dependency analsyis, only inputs/outputs
                self.info = dependency.analyze_bindings(self.model.root)
            return self.info

    #--------------------------------------------------
    # IR handlers
    #--------------------------------------------------

    def handle(self, task: ir.Task, ctx: CDC.CDCContext):
        # currently we only extract if it's a sequence of Logicals, but we could in the
        # future support other intermediate nodes
        if isinstance(task, ir.Logical):
            return self.handle_logical(task, ctx)
        elif isinstance(task, ir.Not):
            return self.handle_not(task, ctx)
        else:
            return task

    def handle_logical(self, task: ir.Logical, ctx: CDC.CDCContext):

        wide_cdc_table_lookups = ordered_set()
        for child in task.body:
            if isinstance(child, ir.Lookup) and bt.from_cdc_annotation in child.relation.annotations:
                wide_cdc_table_lookups.add(child)

        # optimization to avoid creating a frame if unnecessary
        if not wide_cdc_table_lookups:
            # no need to analyze dependencies, just handle children recursively and
            # reconstruct the logical
            body:OrderedSet[ir.Task] = ordered_set()
            for child in task.body:
                body.add(self.handle(child, ctx))
            return ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations)

        # ensure function annotation is in the model
        # TODO: need to revisit this to add @function annotations only when we really need them
        # ctx.analysis_ctx.relations.append(rel_bt.function)

        # process the children
        body:OrderedSet[ir.Task] = ordered_set()

        # find variables required by the other tasks
        required_vars = ordered_set()
        for child in task.body:
            if child not in wide_cdc_table_lookups:
                required_vars.update(ctx.get_info().task_inputs(child))
                #
                # Vars used in atoms (outputs) may be column variables:
                # e.g.:
                #       Transaction.type(source.COLUMN)
                #    or
                #       Transaction.type = source.COLUMN
                #
                # Neither of these mark their vars as inputs, so we pull from the outputs as well.
                # The only exception is RowId, since there's no need to join on `METADATA$KEY` if
                # a non-table atom has `RowId` in keys.
                #
                output_vars = ctx.get_info().task_outputs(child) or []
                output_vars = [var for var in output_vars if not types.matches(var.type, types.RowId)]
                required_vars.update(output_vars)

        # We must return anything that's hoisted
        required_vars.update(helpers.hoisted_vars(task.hoisted))

        # rewrite the cdc table lookup into lookups for each required variable
        for child in task.body:
            if child in wide_cdc_table_lookups:
                assert isinstance(child, ir.Lookup)
                wide_relation = child.relation
                properties = required_vars & ctx.get_info().task_outputs(child)
                if properties:
                    assert isinstance(child.args[0], ir.Var) and types.matches(child.args[0].type, types.RowId)
                    row_id = child.args[0]
                    for property in properties:
                        if types.matches(property.type, types.RowId) and len(properties) > 1:
                            continue

                        relation = self._get_property_cdc_relation(wide_relation, property, ctx)
                        field_name = ir.Literal(types.Symbol, property.name)
                        if types.matches(property.type, types.RowId):
                            field_name = ir.Literal(types.Symbol, "METADATA$KEY")
                            property = ir.Var(type=types.RowId, name=property.name)

                            # METADATA$KEY is unary
                            relation = f.relation(
                                relation.name,
                                [f.field("symbol", types.Symbol), f.field("row_id", types.RowId)],
                                annos=[*relation.annotations],
                            )
                            body.add(ir.Lookup(
                                task.engine,
                                relation,
                                tuple([field_name, row_id])
                            ))
                        else:
                            body.add(ir.Lookup(
                                task.engine,
                                relation,
                                tuple([field_name, row_id, property])
                            ))

            # handle non cdc table children, adding @function to the updates
            for child in task.body:
                if child not in wide_cdc_table_lookups:
                    body.add(self.handle(child, ctx))

                    # TODO: need to revisit this to add @function annotations only when we really need them
                    # replacement = self.handle(child, ctx)
                    # if isinstance(replacement, ir.Update):
                    #     if len(replacement.args) == 1:
                    #         body.add(replacement)
                    #     else:
                    #         body.add(replacement.reconstruct(
                    #             replacement.engine,
                    #             replacement.relation,
                    #             replacement.args,
                    #             replacement.effect,
                    #             replacement.annotations | [rel_bt.function_annotation]
                    #         ))
                    # else:
                    #     body.add(replacement)

        return ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations)

    def handle_not(self, not_task: ir.Not, ctx: CDC.CDCContext):
        sub_task = self.handle(not_task.task, ctx)
        return ir.Not(not_task.engine, sub_task, not_task.annotations)

    def _get_property_cdc_relation(self, wide_cdc_relation: ir.Relation, property: ir.Var, ctx: CDC.CDCContext):
        """
        Get the relation that represents this property var in this wide_cdc_relation. If the
        relation is not yet available in the context, this method will create and register it.
        """
        relation_name = helpers.sanitize(wide_cdc_relation.name).replace("-", "_")
        key = (relation_name, property.name)
        if key not in ctx.cdc_relations:
            # the property relation is overloaded for all properties of the same wide cdc relation, so they have
            # the same name, but potentially a different type in the value column; also note that they are
            # annotated as external to avoid renaming.
            relation = f.relation(
                relation_name,
                [f.field("symbol", types.Symbol), f.field("row_id", types.Number), f.field("value", property.type)],
                annos=[bt.external_annotation]
            )
            ctx.cdc_relations[key] = relation
            ctx.rewrite_ctx.relations.append(relation)
        return ctx.cdc_relations[key]



"""
Engine state handlers for auto_create_engine methods using Strategy Pattern.

Each state handler encapsulates the logic for handling a specific engine state
(PENDING, SUSPENDED, READY, GONE, or None/missing). Handlers are separated
for sync and async modes since they have different behaviors.
"""
from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, Any, cast, Union

from ....errors import EngineNotFoundException, EngineProvisioningFailed, EngineResumeFailed, EngineSizeMismatchWarning
from ....tools.cli_controls import Spinner
from ...util import poll_with_specified_overhead

if TYPE_CHECKING:
    from .snowflake import Resources
    from ...types import EngineState
    EngineDict = Union[EngineState, Dict[str, Any]]
else:
    EngineDict = Dict[str, Any]


@dataclass
class EngineContext:
    """Context for engine state handling."""
    engine_name: str
    engine_size: str | None
    headers: Dict | None
    requested_size: str | None  # Size explicitly requested by user
    spinner: Spinner | None = None  # For async mode UI updates
    span: Any = None  # For sync mode debugging span


class EngineStateHandler(ABC):
    """Base class for engine state handlers using Strategy Pattern."""

    @abstractmethod
    def handles_state(self, state: str | None) -> bool:
        """Check if this handler can process the given engine state."""
        pass

    @abstractmethod
    def handle(self, engine: EngineDict | None, context: EngineContext, resources: 'Resources') -> EngineDict | None:
        """
        Handle the engine state and return updated engine dict or None.

        Returns:
            - Updated engine dict if engine should remain
            - None if engine should be deleted/recreated
        """
        pass


# ============================================================================
# Sync Mode Handlers
# ============================================================================

class SyncPendingStateHandler(EngineStateHandler):
    """Handle PENDING state in sync mode - poll until ready."""

    def handles_state(self, state: str | None) -> bool:
        return state == "PENDING"

    def handle(self, engine: EngineDict | None, context: EngineContext, resources: 'Resources') -> EngineDict | None:
        if not engine:
            return None

        # Warn if requested size doesn't match pending engine size
        if context.requested_size is not None and engine.get("size") != context.requested_size:
            existing_size = engine.get("size") or ""
            EngineSizeMismatchWarning(context.engine_name, existing_size, context.requested_size)

        # Poll until engine is ready
        with Spinner(
            "Waiting for engine to be initialized",
            "Engine ready",
        ):
            poll_with_specified_overhead(
                lambda: resources.is_engine_ready(context.engine_name),
                overhead_rate=0.1,
                max_delay=0.5,
                timeout=900
            )

        # Return updated engine (should be READY now)
        updated_engine = resources.get_engine(context.engine_name)
        return cast(EngineDict, updated_engine) if updated_engine else None


class SyncSuspendedStateHandler(EngineStateHandler):
    """Handle SUSPENDED state in sync mode - resume and poll until ready."""

    def handles_state(self, state: str | None) -> bool:
        return state == "SUSPENDED"

    def handle(self, engine: EngineDict | None, context: EngineContext, resources: 'Resources') -> EngineDict | None:
        if not engine:
            return None

        with Spinner(
            f"Resuming engine '{context.engine_name}'",
            f"Engine '{context.engine_name}' resumed",
            f"Failed to resume engine '{context.engine_name}'"
        ):
            try:
                resources.resume_engine_async(context.engine_name, headers=context.headers)
                poll_with_specified_overhead(
                    lambda: resources.is_engine_ready(context.engine_name),
                    overhead_rate=0.1,
                    max_delay=0.5,
                    timeout=900
                )
            except Exception:
                raise EngineResumeFailed(context.engine_name)

        # Return updated engine (should be READY now)
        updated_engine = resources.get_engine(context.engine_name)
        return cast(EngineDict, updated_engine) if updated_engine else None


class SyncReadyStateHandler(EngineStateHandler):
    """Handle READY state in sync mode - set active and return."""

    def handles_state(self, state: str | None) -> bool:
        return state == "READY"

    def handle(self, engine: EngineDict | None, context: EngineContext, resources: 'Resources') -> EngineDict | None:
        if not engine:
            return None

        # Warn if requested size doesn't match ready engine size
        if context.requested_size is not None and engine.get("size") != context.requested_size:
            existing_size = engine.get("size") or ""
            EngineSizeMismatchWarning(context.engine_name, existing_size, context.requested_size)

        # Cast to EngineState for _set_active_engine
        if TYPE_CHECKING:
            from ...types import EngineState
            resources._set_active_engine(cast(EngineState, engine))
        else:
            resources._set_active_engine(engine)  # type: ignore[arg-type]
        return engine


class SyncGoneStateHandler(EngineStateHandler):
    """Handle GONE state in sync mode - delete and return None to trigger recreation."""

    def handles_state(self, state: str | None) -> bool:
        return state == "GONE"

    def handle(self, engine: EngineDict | None, context: EngineContext, resources: 'Resources') -> EngineDict | None:
        if not engine:
            return None

        try:
            # "Gone" is abnormal condition when metadata and SF service don't match
            # Therefore, we have to delete the engine and create a new one
            # It could be case that engine is already deleted, so we have to catch the exception
            resources.delete_engine(context.engine_name, headers=context.headers)
            # After deleting the engine, return None so that we can create a new engine
            return None
        except Exception as e:
            # If engine is already deleted, we will get an exception
            # We can ignore this exception and create a new engine
            if isinstance(e, EngineNotFoundException):
                return None
            else:
                raise EngineProvisioningFailed(context.engine_name, e) from e


class SyncMissingEngineHandler(EngineStateHandler):
    """Handle missing engine (None) in sync mode - create synchronously."""

    def handles_state(self, state: str | None) -> bool:
        return state is None

    def handle(self, engine: EngineDict | None, context: EngineContext, resources: 'Resources') -> EngineDict | None:
        # This handler is called when engine doesn't exist
        # Create engine synchronously with spinner
        with Spinner(
            f"Auto-creating engine {context.engine_name}",
            f"Auto-created engine {context.engine_name}",
            "Engine creation failed",
        ):
            resources.create_engine(context.engine_name, size=context.engine_size, headers=context.headers)

        return resources.get_engine(context.engine_name)


# ============================================================================
# Async Mode Handlers
# ============================================================================

class AsyncPendingStateHandler(EngineStateHandler):
    """Handle PENDING state in async mode - just update spinner, don't poll."""

    def handles_state(self, state: str | None) -> bool:
        return state == "PENDING"

    def handle(self, engine: EngineDict | None, context: EngineContext, resources: 'Resources') -> EngineDict | None:
        if not engine:
            return None

        # In async mode, just update spinner - use_index will wait for engine to be ready
        if context.spinner:
            context.spinner.update_messages({
                "finished_message": f"Starting engine {context.engine_name}",
            })

        return engine


class AsyncSuspendedStateHandler(EngineStateHandler):
    """Handle SUSPENDED state in async mode - resume asynchronously."""

    def handles_state(self, state: str | None) -> bool:
        return state == "SUSPENDED"

    def handle(self, engine: EngineDict | None, context: EngineContext, resources: 'Resources') -> EngineDict | None:
        if not engine:
            return None

        if context.spinner:
            context.spinner.update_messages({
                "finished_message": f"Resuming engine {context.engine_name}",
            })

        try:
            resources.resume_engine_async(context.engine_name)
        except Exception:
            raise EngineResumeFailed(context.engine_name)

        return engine


class AsyncReadyStateHandler(EngineStateHandler):
    """Handle READY state in async mode - set active."""

    def handles_state(self, state: str | None) -> bool:
        return state == "READY"

    def handle(self, engine: EngineDict | None, context: EngineContext, resources: 'Resources') -> EngineDict | None:
        if not engine:
            return None

        if context.spinner:
            context.spinner.update_messages({
                "finished_message": f"Engine {context.engine_name} initialized",
            })

        # Cast to EngineState for _set_active_engine
        if TYPE_CHECKING:
            from ...types import EngineState
            resources._set_active_engine(cast(EngineState, engine))
        else:
            resources._set_active_engine(engine)  # type: ignore[arg-type]
        return engine


class AsyncGoneStateHandler(EngineStateHandler):
    """Handle GONE state in async mode - delete and return None to trigger recreation."""

    def handles_state(self, state: str | None) -> bool:
        return state == "GONE"

    def handle(self, engine: EngineDict | None, context: EngineContext, resources: 'Resources') -> EngineDict | None:
        if not engine:
            return None

        if context.spinner:
            context.spinner.update_messages({
                "message": f"Restarting engine {context.engine_name}",
            })

        try:
            # "Gone" is abnormal condition when metadata and SF service don't match
            # Therefore, we have to delete the engine and create a new one
            # It could be case that engine is already deleted, so we have to catch the exception
            # Set it to None so that we can create a new engine asynchronously
            resources.delete_engine(context.engine_name)
            return None
        except Exception as e:
            # If engine is already deleted, we will get an exception
            # We can ignore this exception and create a new engine asynchronously
            if isinstance(e, EngineNotFoundException):
                return None
            else:
                raise EngineProvisioningFailed(context.engine_name, e) from e


class AsyncMissingEngineHandler(EngineStateHandler):
    """Handle missing engine (None) in async mode - create asynchronously."""

    def handles_state(self, state: str | None) -> bool:
        return state is None

    def handle(self, engine: EngineDict | None, context: EngineContext, resources: 'Resources') -> EngineDict | None:
        # This handler is called when engine doesn't exist
        # Create engine asynchronously
        resources.create_engine_async(context.engine_name, size=context.engine_size)

        if context.spinner:
            context.spinner.update_messages({
                "finished_message": f"Starting engine {context.engine_name}...",
            })
        return None  # Engine is being created asynchronously




import warnings

from v0.relationalai.semantics.std.decimals import decimal, parse_decimal

__all__ = ["decimal", "parse_decimal"]

warnings.warn(
    "relationalai.early_access.builder.std.decimals is deprecated, "
    "Please migrate to relationalai.semantics.std.decimals",
    DeprecationWarning,
    stacklevel=2,
)


from __future__ import annotations
from v0.relationalai.semantics.internal import snowflake as sf, internal as b
from v0.relationalai.early_access.dsl.bindings.common import BindableColumn, AbstractBindableTable


#=
# Bindable classes and interfaces.
#=

class BindableSnowflakeColumn(BindableColumn, sf.Column):
    _metadata: b.Field

    def __init__(self, metadata: b.Field, table: 'SnowflakeTable', model):
        col_name = metadata.name
        col_type = b.field_to_type(model.qb_model(), metadata)
        sf.Column.__init__(self, table, col_name)
        BindableColumn.__init__(self, col_name, col_type, table, model)
        self._metadata = metadata

    @property
    def metadata(self):
        return self._metadata

    def __repr__(self):
        return f"Snowflake:{super().__repr__()}"


class SnowflakeTable(AbstractBindableTable[BindableSnowflakeColumn], sf.Table):

    def __init__(self, fqn: str, model, schema:dict[str, str|b.Concept]|None=None):
        sf.Table.__init__(self, fqn, schema=schema)
        AbstractBindableTable.__init__(self, fqn, model, set())
        self._initialize(model)

    def _initialize(self, model):
        self._lazy_init()
        self._model = model
        schema_info = self._schemas[(self._database, self._schema)]
        table_info = schema_info.tables[self._table]
        self._cols = {field.name: BindableSnowflakeColumn(field, self, model) for field in table_info.fields}
        self._generate_declare()

    def _generate_declare(self):
        src = f"declare {self.physical_name()}"
        self._model.qb_model().define(b.RawSource('rel', src))

    def _to_type(self) -> b.Concept:
        return self._concept

    def _to_keys(self) -> list[b.Ref]:
        return [self._rel._field_refs[0]]

    def _compile_lookup(self, compiler:b.Compiler, ctx:b.CompilerContext):
        sf.Table._used_sources.add(self)
        compiler.lookup(self._rel, ctx)
        return compiler.lookup(b.RelationshipFieldRef(None, self._rel, 0), ctx)

    def __str__(self):
        # returns the name of the table, as well as the columns and their types
        return f"SnowflakeTable({self.physical_name()}"

    def physical_name(self):
        # physical relation name is always in the form of `{database}_{schema}_{table}
        return f"{self._fqn.lower()}".replace('.', '_')



from __future__ import annotations


from . import StringValue, IntegerValue, strings, _function_not_implemented
from ..frontend.base import Library, Expression, Field, Variable
from ..frontend.core import Number, String, Integer
from enum import Enum

# the front-end library object
library = Library("re")

#--------------------------------------------------
# Relationships
#--------------------------------------------------

_regex_match_all = library.Relation("regex_match_all", [Field.input("regex", String), Field.input("value", String), Field("pos", Integer), Field("match", String)])
# _regex_replace = library.Relation("regex_replace", [Field.input("regex", String), Field.input("replacement", String), Field.input("value", String), Field("result", String)])

# _regex_escape = library.Relation("regex_escape", [Field.input("regex", String), Field("result", String)])
# _capture_group_by_index = library.Relation("capture_group_by_index", [Field.input("regex", String), Field.input("string", String), Field.input("pos", Integer), Field.input("index", Integer), Field("result", String)])
# _capture_group_by_name = library.Relation("capture_group_by_name", [Field.input("regex", String), Field.input("string", String), Field.input("pos", Integer), Field.input("name", String), Field("result", String)])


#--------------------------------------------------
# Operations
#--------------------------------------------------

#  match is ^{REGEX} , search is {REGEX}, and fullmatch is ^{REGEX}$

def match(regex: StringValue, value: StringValue) -> RegexMatch:
    """ Check if the regex matches the value from the start. """
    # REGEXP_LIKE
    return RegexMatch(regex, value, type=RegexMatchType.MATCH)

def search(regex: StringValue, value: StringValue, pos: IntegerValue = 0) -> RegexMatch:
    _function_not_implemented("re.search")
    return RegexMatch(regex, value, pos, type=RegexMatchType.SEARCH)

def fullmatch(regex: StringValue, value: StringValue, pos: IntegerValue = 0) -> RegexMatch:
    """ Check if the regex matches the entire value starting at the given position. """
    return RegexMatch(regex, value, pos, type=RegexMatchType.FULLMATCH)

def findall(regex: StringValue, value: StringValue) -> tuple[Variable, Variable]:
    """ Find all non-overlapping matches of the regex in the value. """
    _function_not_implemented("re.findall")
    # exp = _regex_match_all(regex, value)
    # ix, match = exp._arg_ref(2), exp._arg_ref(3)
    # rank = i.rank(i.asc(ix, match))
    # return rank, match
    raise

def sub(regex: StringValue, repl: StringValue, value: StringValue):
    """ Replace occurrences of the regex in the value with the replacement string. """
    _function_not_implemented("re.sub")
    # return _regex_replace(regex, repl, value)

class RegexMatchType(Enum):
    MATCH = "match"
    SEARCH = "search"
    FULLMATCH = "fullmatch"

class RegexMatch(Expression):

    def __init__(self, regex: StringValue, value: StringValue, pos: IntegerValue = 0, type=RegexMatchType.MATCH):
        if type == RegexMatchType.FULLMATCH:
            # fullmatch: ^{REGEX}$
            self.regex = strings.concat(regex, "$")
        else:
            self.regex = regex
        self.value = value
        # pos is the 0-based index in value where we start matching
        self.pos = pos
        # return value is the matched string
        self.match = String.ref("match")
        super().__init__(_regex_match_all, (self.regex, value, self.pos, self.match))

    def start(self) -> IntegerValue:
        return self.pos

    def end(self) -> IntegerValue:
        return strings.len(self.match) + self.pos - 1

    def span(self) -> tuple[IntegerValue, IntegerValue]:
        return self.start(), self.end()

    def group(self, index: IntegerValue = 0) -> Variable:
        _function_not_implemented("re.RegexMatch.group")
        raise

    def group_by_name(self, name: StringValue) -> Variable:
        _function_not_implemented("re.RegexMatch.group_by_name")
        raise

#--------------------------------------------------
# Helpers
#--------------------------------------------------

# def _regex_match_all(regex: StringValue, string: StringValue, pos: IntegerValue|None = None) -> i.Expression:
#     if pos is None:
#         pos = i.Int64.ref()
#     return _make_expr("regex_match_all", regex, string, pos, i.StringValue.ref())



from v0.relationalai.semantics import Model, Integer, define, select
from v0.relationalai.experimental.paths.graph import Graph
from v0.relationalai.experimental.paths.path_algorithms.usp import compute_usp, compute_nsp


# First test for usp with grid graph
model_grid = Model("test_usp_nsp_multiple grid", dry_run=False)

grid1 = Graph.construct_grid(model_grid, 3)

source1_1 = grid1.Node.new(row = 1, col = 1)
source1_2 = grid1.Node.new(row = 1, col = 2)
target1_1 = grid1.Node.new(row = 3, col = 2)
target1_2 = grid1.Node.new(row = 3, col = 3)

Source1 = model_grid.Concept("Source1", extends=[grid1.Node])
Target1 = model_grid.Concept("Target1", extends=[grid1.Node])

define(Source1(source1_1))
define(Source1(source1_2))
define(Target1(target1_1))
define(Target1(target1_2))

u, v = grid1.Node.ref(), grid1.Node.ref()
n = Integer.ref()

usp1, _ = compute_usp(grid1, Source1, Target1)
iter_usp1 = select(u.row, u.col, v.row, v.col).where(usp1(u, v)).to_df()
set_usp1 = set(row for row in iter_usp1.itertuples(index = False, name = None))

expected_usp1 = {
    (1, 2, 2, 2), (2, 2, 3, 2)
}

assert set_usp1 == expected_usp1


# Second test for usp with grid graph
Target2 = model_grid.Concept("Target2", extends=[grid1.Node])

define(Target2(target1_2))

u, v = grid1.Node.ref(), grid1.Node.ref()
n = Integer.ref()

usp2, _ = compute_usp(grid1, Source1, Target2)
iter_usp2 = select(u.row, u.col, v.row, v.col).where(usp2(u, v)).to_df()
set_usp2 = set(row for row in iter_usp2.itertuples(index = False, name = None))

expected_usp2 = {
    (1, 2, 2, 2), (2, 2, 3, 2),
    (1, 3, 2, 3), (2, 3, 3, 3),
    (1, 2, 1, 3), (2, 2, 2, 3), (3, 2, 3, 3)
}

assert set_usp2 == expected_usp2


# First test for usp with diamond graph
model_diamond = Model("test_usp_nsp_multiple diamond", dry_run=False)

diamond3 = Graph.construct_diamond(model_diamond, 3)

du, dv = diamond3.Node.ref(), diamond3.Node.ref()

source3_1 = diamond3.Node.new(id = 2)
source3_2 = diamond3.Node.new(id = 3)
target3_1 = diamond3.Node.new(id = 8)
target3_2 = diamond3.Node.new(id = 9)

Source3 = model_diamond.Concept("Source3", extends=[diamond3.Node])
Target3 = model_diamond.Concept("Target3", extends=[diamond3.Node])

define(Source3(source3_1))
define(Source3(source3_2))
define(Target3(target3_1))
define(Target3(target3_2))

usp3, _ = compute_usp(diamond3, Source3, Target3)
iter_usp3 = select(du.id, dv.id).where(usp3(du, dv)).to_df()
set_usp3 = set(row for row in iter_usp3.itertuples(index = False, name = None))

expected_usp3 = {
    (2, 4), (3, 4), (4, 5), (4, 6), (5, 7), (6, 7), (7, 9), (7, 8)
}

assert set_usp3 == expected_usp3


# First test for nsp with grid graph
nsp1 = compute_nsp(grid1, Source1, Target1)
iter_nsp1 = select(u.row, u.col, n).where(nsp1(u, n)).to_df()
set_nsp1 = set(row for row in iter_nsp1.itertuples(index = False, name = None))

expected_nsp1 = {
    (1, 2, 1), (2, 2, 1), (3, 2, 1)
}

assert set_nsp1 == expected_nsp1


# Second test for nsp with diamond graph
model_diamond_2 = Model("test_usp_nsp_multiple diamond_2", dry_run=False)

diamond4 = Graph.construct_diamond(model_diamond_2, 5)

dw = diamond4.Node.ref()

source4_1 = diamond4.Node.new(id = 4)
source4_2 = diamond4.Node.new(id = 5)
source4_3 = diamond4.Node.new(id = 7)
target4_1 = diamond4.Node.new(id = 7)
target4_2 = diamond4.Node.new(id = 9)
target4_3 = diamond4.Node.new(id = 10)

Source4 = model_diamond_2.Concept("Source4", extends=[diamond4.Node])
Target4 = model_diamond_2.Concept("Target4", extends=[diamond4.Node])

define(Source4(source4_1))
define(Source4(source4_2))
define(Source4(source4_3))
define(Target4(target4_1))
define(Target4(target4_2))
define(Target4(target4_3))

nsp4 = compute_nsp(diamond4, Source4, Target4)
iter_nsp4 = select(dw.id, n).where(nsp4(dw, n)).to_df()
set_nsp4 = set(row for row in iter_nsp4.itertuples(index = False, name = None))

expected_nsp4 = {
    (7, 1)
}

assert set_nsp4 == expected_nsp4


# First test with cyclic graph
model_cyclic = Model("test_usp_nsp_multiple cyclic", dry_run=False)

cyclic5 = Graph.from_edge_list(model_cyclic, [(1, 2), (2, 3), (3, 4), (2, 5), (5, 2), (5, 6)])

cu, cv = cyclic5.Node.ref(), cyclic5.Node.ref()

source5 = cyclic5.Node.new(id = 1)
target5_1 = cyclic5.Node.new(id = 4)
target5_2 = cyclic5.Node.new(id = 6)

Source5 = model_cyclic.Concept("Source5", extends=[cyclic5.Node])
Target5 = model_cyclic.Concept("Target5", extends=[cyclic5.Node])

define(Source5(source5))
define(Target5(target5_1))
define(Target5(target5_2))

usp5, _ = compute_usp(cyclic5, Source5, Target5)
iter_usp5 = select(cu.id, cv.id).where(usp5(cu, cv)).to_df()
set_usp5 = set(row for row in iter_usp5.itertuples(index = False, name = None))

nsp5 = compute_nsp(cyclic5, Source5, Target5)
iter_nsp5 = select(cu.id, n).where(nsp5(cu, n)).to_df()
set_nsp5 = set(row for row in iter_nsp5.itertuples(index = False, name = None))

expected_usp5 = {
    (1, 2), (2, 3), (3, 4), (2, 5), (5, 6)
}

expected_nsp5 = {
    (1, 2), (2, 2), (3, 1), (4, 1), (5, 1), (6, 1)
}

assert (
    set_usp5 == expected_usp5 and
    set_nsp5 == expected_nsp5
)


# First test for usp and nsp with undirected diamond graph
model_und = Model("test_usp_nsp_multiple und", dry_run=False)

diamond6 = Graph.construct_diamond(model_und, 5, "undirected")

uu, uv = diamond6.Node.ref(), diamond6.Node.ref()

source6_1 = diamond6.Node.new(id = 7)
source6_2 = diamond6.Node.new(id = 10)
target6_1 = diamond6.Node.new(id = 4)
target6_2 = diamond6.Node.new(id = 13)

Source6 = model_und.Concept("Source6", extends=[diamond6.Node])
Target6 = model_und.Concept("Target6", extends=[diamond6.Node])

define(Source6(source6_1))
define(Source6(source6_2))
define(Target6(target6_1))
define(Target6(target6_2))

usp6, _ = compute_usp(diamond6, Source6, Target6)
iter_usp6 = select(uu.id, uv.id).where(usp6(uu, uv)).to_df()
set_usp6 = set(row for row in iter_usp6.itertuples(index = False, name = None))

nsp6 = compute_nsp(diamond6, Source6, Target6)
iter_nsp6 = select(uu.id, n).where(nsp6(uu, n)).to_df()
set_nsp6 = set(row for row in iter_nsp6.itertuples(index = False, name = None))

expected_usp6 = {
    (7, 6), (7, 5), (6, 4), (5, 4),
    (10, 11), (10, 12), (11, 13), (12, 13)
}

expected_nsp6 = {
    (4, 1), (5, 1), (6, 1), (7, 2),
    (13, 1), (12, 1), (11, 1), (10, 2)
}

assert (
    set_usp6 == expected_usp6 and
    set_nsp6 == expected_nsp6
)


# First test for usp and nsp with grid graph
target7 = grid1.Node.new(row = 4, col = 4)

Target7 = model_grid.Concept("Target7", extends=[grid1.Node])

usp7, _ = compute_usp(grid1, Source1, Target7)
nsp7 = compute_nsp(grid1, Source1, Target7)

assert usp7.to_df().empty and nsp7.to_df().empty



from __future__ import annotations
from typing import Tuple

from v0.relationalai.semantics.metamodel import builtins, ir, factory as f, types, visitor
from v0.relationalai.semantics.metamodel.compiler import Pass, group_tasks
from v0.relationalai.semantics.metamodel.util import OrderedSet
from v0.relationalai.semantics.metamodel.util import FrozenOrderedSet
from v0.relationalai.semantics.metamodel.typer.typer import is_primitive

class FormatOutputs(Pass):
    def __init__(self, handle_outputs: bool=True):
        super().__init__()
        self._handle_outputs = handle_outputs

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        wide_outputs = options.get("wide_outputs", False)
        return self.OutputRewriter(wide_outputs).walk(model)

    class OutputRewriter(visitor.Rewriter):
        def __init__(self, wide_outputs: bool = False):
            super().__init__()
            self.wide_outputs = wide_outputs

        def handle_logical(self, node: ir.Logical, parent: ir.Node):
            # Rewrite children first
            node = super().handle_logical(node, parent)

            groups = group_tasks(node.body, {
                "outputs": ir.Output,
            })

            # If no outputs, return as is
            if not groups["outputs"]:
                return node

            return adjust_outputs(node, groups["outputs"], self.wide_outputs)

#--------------------------------------------------
# GNF vs wide output support
#--------------------------------------------------
def adjust_outputs(task: ir.Logical, outputs: OrderedSet[ir.Task], wide_outputs: bool = False):

    body = list(task.body)

    # For wide outputs, only adjust the output task to include the keys.
    if wide_outputs:
        for output in outputs:
            assert(isinstance(output, ir.Output))
            if output.keys:
                body.remove(output)
                body.append(rewrite_wide_output(output))
        return ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations)

    # For GNF outputs we need to generate a rule for each "column" in the output
    else:
        # First split outputs in potentially multiple outputs, one for each "column"
        for output in outputs:
            assert(isinstance(output, ir.Output))
            if output.keys:
                # Remove the original output. This is replaced by per-column outputs below
                body.remove(output)

                is_export = builtins.export_annotation in output.annotations

                # Generate an output for each "column"
                # output looks like def output(:cols, :col000, key0, key1, value):
                original_cols = OrderedSet()
                for idx, alias in enumerate(output.aliases):
                    # Skip None values which are used as a placeholder for missing values
                    if alias[1] is None:
                        continue
                    original_cols.add(alias[1])
                    body.extend(_generate_output_column(output, idx, alias, is_export))

                idx = len(output.aliases)
                for key in output.keys:
                    if key not in original_cols:
                        body.extend(_generate_output_column(output, idx, (key.name, key), is_export))
                        idx += 1

        return ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations)

# TODO: return non list?
def _generate_output_column(output: ir.Output, idx: int, alias: tuple[str, ir.Value], is_export: bool):
    if not output.keys:
        return [output]

    aliases = [("cols", f.literal("cols", types.Symbol))] if not is_export else []
    aliases.append(("col", f.literal(f"col{idx:03}", types.Symbol)))

    # Append all keys at the start
    for k in output.keys:
        aliases.append((f"key_{k.name}_{idx}", k))

    if (is_export and
        isinstance(alias[1], ir.Var) and
        (not is_primitive(alias[1].type) or alias[1].type == types.Hash)):

        uuid = f.var(f"{alias[0]}_{idx}_uuid", types.String)
        aliases.append((uuid.name, uuid))

        return [
            ir.Lookup(None, builtins.uuid_to_string, (alias[1], uuid)),
            ir.Output(
                output.engine,
                FrozenOrderedSet.from_iterable(aliases),
                output.keys,
                output.annotations
            )
        ]
    else:
        aliases.append(alias)

        return [
            ir.Output(
                output.engine,
                FrozenOrderedSet.from_iterable(aliases),
                output.keys,
                output.annotations
            )
        ]

def rewrite_wide_output(output: ir.Output):
    assert(output.keys)

    # Only append keys that are not already in the output
    suffix_keys = []
    for key in output.keys:
        if all([val is not key for _, val in output.aliases]):
            suffix_keys.append(key)

    aliases: OrderedSet[Tuple[str, ir.Value]] = OrderedSet()

    # Add the remaining args, unless it is already a key
    for name, val in output.aliases:
        if not isinstance(val, ir.Var) or val not in suffix_keys:
            aliases.add((name, val))

    # Add the keys to the output
    for key in suffix_keys:
        aliases.add((key.name, key))

    # TODO - we are assuming that the Rel compiler will translate nullable lookups
    # properly, returning a `Missing` if necessary, like this:
    # (nested_192(_adult, _adult_name) or (not nested_192(_adult, _) and _adult_name = Missing)) and
    return ir.Output(
        output.engine,
        aliases.frozen(),
        output.keys,
        output.annotations
    )

    # TODO: in the rel compiler, see if we can do this outer join
    # 1. number of keys
    # 2. each relation
    # 3. each variable, starting with the keys
    # 4. tag output with @arrow

    # @arrow def output(_book, _book_title, _author_name):
    #   rel_primitive_outer_join(#1, book_title, author_name, _book, _book_title, _author_name)
    # def output(p, n, c):
    #     rel_primitive_outer_join(#1, name, coolness, p, n, c)



import warnings

from v0.relationalai.semantics.lqp import ir, model2lqp, result_helpers, utils

__all__ = ['ir', 'model2lqp', 'result_helpers', 'utils']

warnings.warn(
    "relationalai.early_access.lqp.* is deprecated. "
    "Please migrate to relationalai.semantics.lqp.*",
    DeprecationWarning,
    stacklevel=2,
)



from dataclasses import dataclass, field, fields, is_dataclass
from functools import lru_cache
from typing import Tuple, Any

def is_dataclass_instance(obj) -> bool:
    """Checks if the object is a dataclass instance, excluding the class itself."""
    return is_dataclass(obj) and not isinstance(obj, type)

def print_tree(node, field_name=None, indent=0, hide_fields=[], seen=set()):
    """Recursively print a dataclass tree structure.

    Args:
        node: The dataclass instance or value to print.
        field_name: Optional name of the field (for nested calls).
        indent: Current indentation level.
        hide_fields: List of field names to hide in the output."""
    if id(node) in seen:
        _indent_print(indent, "<recursion>", field_name)
        return
    seen.add(id(node))
    if is_dataclass_instance(node):
        _indent_print(indent, f"{node.__class__.__name__}:", field_name)
        next = indent + 2
        for field in fields(node):
            if field.name not in hide_fields:
                print_tree(getattr(node, field.name), field.name, next, hide_fields)
    elif isinstance(node, list) or isinstance(node, tuple):
        # just ignore sequences that are empty
        if node:
            _indent_print(indent, f"[", field_name)
            next = indent + 2
            for item in node:
                print_tree(item, None, next, hide_fields)
            _indent_print(indent, f"]")
    else:
        _indent_print(indent, str(node), field_name)


def _indent_print(indent, value, field_name=None):
    if field_name:
        print(" " * indent + f"{field_name}: {value}")
    else:
        print(" " * indent + value)



from __future__ import annotations

from dataclasses import dataclass
from typing import  Iterable, Sequence as seq

from relationalai.semantics.metamodel import metamodel as mm
from relationalai.semantics.metamodel.builtins import builtins as b
from relationalai.semantics.metamodel.rewriter import NO_WALK, Walker
from relationalai.util.structures import OrderedSet


class ContainerWalker(Walker):
    """ Walker that redirects all container nodes to a common handler. """

    def enter_container(self, container: mm.Container, children: seq[mm.Node]) -> object |  None:
        return None

    def exit_container(self, container: mm.Container, children: seq[mm.Node]) -> object |  None:
        return container

    #------------------------------------------------------
    # Redirect containers to common handlers
    #------------------------------------------------------

    def enter_logical(self, logical: mm.Logical):
        return self.enter_container(logical, logical.body)

    def logical(self, logical: mm.Logical):
        return self.exit_container(logical, logical.body)

    def enter_sequence(self, sequence: mm.Sequence):
        return self.enter_container(sequence, sequence.tasks)

    def sequence(self, sequence: mm.Sequence):
        return self.exit_container(sequence, sequence.tasks)

    def enter_union(self, union: mm.Union):
        return self.enter_container(union, union.tasks)

    def union(self, union: mm.Union):
        return self.exit_container(union, union.tasks)

    def enter_match(self, match: mm.Match):
        return self.enter_container(match, match.tasks)

    def match(self, match: mm.Match):
        return self.exit_container(match, match.tasks)

    def enter_until(self, until: mm.Until):
        return self.enter_container(until, [until.check, until.body])

    def until(self, until: mm.Until):
        self.exit_container(until, [until.check, until.body])

    def enter_wait(self, wait: mm.Wait):
        return self.enter_container(wait, [wait.check])

    def wait(self, wait: mm.Wait):
        return self.exit_container(wait, [wait.check])

    def enter_loop(self, loop: mm.Loop):
        return self.enter_container(loop, [loop.body])

    def loop(self, loop: mm.Loop):
        return self.exit_container(loop, [loop.body])

    def enter_require(self, require: mm.Require):
        return self.enter_container(require, [require.domain, require.check])

    def require(self, require: mm.Require):
        return self.exit_container(require, [require.domain, require.check])

    def enter_not(self, not_: mm.Not):
        return self.enter_container(not_, [not_.task])

    def not_(self, not_: mm.Not):
        return self.exit_container(not_, [not_.task])

    def enter_exists(self, exists: mm.Exists):
        return self.enter_container(exists, [exists.task])

    def exists(self, exists: mm.Exists):
        return self.exit_container(exists, [exists.task])


class VarFinder(ContainerWalker):
    """ Find all variables used in a set of nodes. """

    def enter_container(self, container, children):
         if self.shallow:
            return NO_WALK
         return None

    def find_vars(self, node:mm.Node|Iterable[mm.Node], positive_only=False, shallow=False) -> OrderedSet[mm.Var]:
        self.vars:OrderedSet[mm.Var] = OrderedSet()
        self.positive_only = positive_only
        self.shallow = shallow
        self(node)
        return self.vars

    def enter_not(self, not_:mm.Not):
        if self.positive_only:
            return NO_WALK

    def var(self, node:mm.Var):
        self.vars.add(node)

#------------------------------------------------------
# Helper functions shared across shim modules
#------------------------------------------------------

def is_output_update(u: mm.Task) -> bool:
    return (
        isinstance(u, mm.Update) and
        len(u.args) > 0 and
        isinstance(u.args[0], mm.Var) and
        isinstance(u.args[0].type, mm.Table) and
        (u.relation is u.args[0].type or u.relation in u.args[0].type.columns)
        # In v0 all outputs, including exports, end up as v0.Output so we actually
        # don't want to limit this to just arrow results
        # u.args[0].type.uri.startswith("dataframe://")
    )

def is_main_output(u: mm.Task) -> bool:
    """ This task is an update to dataframe, and represents the main output (it has only keys, no values)."""
    return isinstance(u, mm.Update) and is_output_update(u) and isinstance(u.relation, mm.Table)



from __future__ import annotations
from typing import Union, Literal
import datetime as dt

from v0.relationalai.semantics.internal import internal as b
from .std import _DateTime, _Date, _Integer, _String, _make_expr
from .. import std

class ISO:
    DATE = "yyyy-mm-dd"
    HOURS = "yyyy-mm-ddTHH"
    HOURS_TZ = "yyyy-mm-ddTHHz"
    MINUTES = "yyyy-mm-ddTHH:MM"
    MINUTES_TZ = "yyyy-mm-ddTHH:MMz"
    SECONDS = "yyyy-mm-ddTHH:MM:SS"
    SECONDS_TZ = "yyyy-mm-ddTHH:MM:SSz"
    MILLIS = "yyyy-mm-ddTHH:MM:SS.s"
    MILLIS_TZ = "yyyy-mm-ddTHH:MM:SS.sz"

#--------------------------------------------------
# Date functions
#--------------------------------------------------

class date:

    def __new__(cls, year: _Integer, month: _Integer, day: _Integer) -> b.Expression:
        return _make_expr("construct_date", std.cast_to_int64(year), std.cast_to_int64(month), std.cast_to_int64(day), b.Date.ref("res"))

    @classmethod
    def year(cls, date: _Date) -> b.Expression:
        return _make_expr("date_year", date, b.Int64.ref("res"))

    @classmethod
    def quarter(cls, date: _Date) -> b.Expression:
        return _make_expr("date_quarter", date, b.Int64.ref("res"))

    @classmethod
    def month(cls, date: _Date) -> b.Expression:
        return _make_expr("date_month", date, b.Int64.ref("res"))

    @classmethod
    def week(cls, date: _Date) -> b.Expression:
        return _make_expr("date_week", date, b.Int64.ref("res"))

    @classmethod
    def day(cls, date: _Date) -> b.Expression:
        return _make_expr("date_day", date, b.Int64.ref("res"))

    @classmethod
    def dayofyear(cls, date: _Date) -> b.Expression:
        return _make_expr("date_dayofyear", date, b.Int64.ref("res"))

    @classmethod
    def isoweekday(cls, date: _Date) -> b.Expression:
        """
        Return the ISO weekday as an integer, where Monday is 1, and Sunday is 7.
        """
        return _make_expr("date_weekday", date, b.Int64.ref("res"))

    @classmethod
    def weekday(cls, date: _Date) -> b.Expression:
        return cls.isoweekday(date) - 1 # Convert ISO weekday (1=Mon..7=Sun) to weekday (0=Mon..6=Sun)

    @classmethod
    def fromordinal(cls, ordinal: _Integer) -> b.Expression:
        # ordinal 1 = '0001-01-01'. Minus 1 day since we can't declare date 0000-00-00
        return cls.add(b.Date(dt.date(1, 1, 1)), days(ordinal - 1))

    @classmethod
    def to_datetime(cls, date: _Date, hour: int = 0, minute: int = 0, second: int = 0, millisecond: int = 0, tz: str = "UTC") -> b.Expression:
        _year = cls.year(date)
        _month = cls.month(date)
        _day = cls.day(date)
        return _make_expr("construct_datetime_ms_tz", _year, _month, _day, hour, minute, second, millisecond, tz, b.DateTime.ref("res"))

    @classmethod
    def format(cls, date: _Date, format: _String) -> b.Expression:
        return _make_expr("date_format", date, format, b.String.ref("res"))

    @classmethod
    def add(cls, date: _Date, period: b.Producer) -> b.Expression:
        return _make_expr("date_add", date, period, b.Date.ref("res"))

    @classmethod
    def subtract(cls, date: _Date, period: b.Producer) -> b.Expression:
        return _make_expr("date_subtract", date, period, b.Date.ref("res"))

    @classmethod
    def range(cls, start: _Date | None = None, end: _Date | None = None, periods: int = 1, freq: Frequency = "D") -> b.Expression:
        """
        Note on date_ranges and datetime_range: The way the computation works is that it first overapproximates the
        number of periods.

        For example, date_range(2025-02-01, 2025-03-01, freq='M') and date_range(2025-02-01, 2025-03-31, freq='M') will
        compute range_end to be ceil(28*1/(365/12))=1 and ceil(58*1/(365/12))=2.

        Then, the computation fetches range_end+1 items into _date, which is the right number in the first case but
        one too many in the second case. That's why a filter end >= _date (or variant of) is applied, to remove any
        extra item. The result is two items in both cases.
        """
        if start is None and end is None:
            raise ValueError("Invalid start/end date for date_range. Must provide at least start date or end date")
        _days = {
            "D": 1,
            "W": 1/7,
            "M": 1/(365/12),
            "Y": 1/365,
        }
        if freq not in _days.keys():
            raise ValueError(f"Frequency '{freq}' is not allowed for date_range. List of allowed frequencies: {list(_days.keys())}")
        date_func = cls.add
        if start is None:
            start = end
            end = None
            date_func = cls.subtract
        assert start is not None
        if end is not None:
            num_days = cls.period_days(start, end)
            if freq in ["W", "M", "Y"]:
                range_end = std.cast(b.Int64, std.math.ceil(num_days * _days[freq]))
            else:
                range_end = num_days
            # date_range is inclusive. add 1 since std.range is exclusive
            ix = std.range(0, range_end + 1, 1)
        else:
            ix = std.range(0, periods, 1)
        _date = date_func(start, _periods[freq](ix))
        if isinstance(end, dt.date) :
            return b.Date(end) >= _date
        elif end is not None:
            return end >= _date
        return _date

    @classmethod
    def period_days(cls, start: _Date, end: _Date) -> b.Expression:
        return _make_expr("dates_period_days", start, end, b.Int64.ref("res"))

    @classmethod
    def fromisoformat(cls, date_string: _String) -> b.Expression:
        return _make_expr("parse_date", date_string, ISO.DATE, b.Date.ref("res"))

#--------------------------------------------------
# DateTime functions
#--------------------------------------------------

class datetime:

    def __new__(cls, year: _Integer, month: _Integer, day: _Integer, hour: _Integer = 0, minute: _Integer = 0,
             second: _Integer = 0, millisecond: _Integer = 0, tz: dt.tzinfo|_String = "UTC") -> b.Expression:
        if isinstance(tz, dt.tzinfo):
            tz = str(tz)
        return _make_expr("construct_datetime_ms_tz", std.cast_to_int64(year), std.cast_to_int64(month),
                          std.cast_to_int64(day), std.cast_to_int64(hour), std.cast_to_int64(minute),
                          std.cast_to_int64(second), std.cast_to_int64(millisecond), tz, b.DateTime.ref("res"))

    @classmethod
    def now(cls) -> b.Expression:
        return _make_expr("datetime_now", b.DateTime.ref("res"))

    @classmethod
    def year(cls, datetime: _DateTime, tz: dt.tzinfo|_String|None = None) -> b.Expression:
        tz = _extract_tz(datetime, tz)
        return _make_expr("datetime_year", datetime, tz, b.Int64.ref("res"))

    @classmethod
    def quarter(cls, datetime: _DateTime,  tz: dt.tzinfo|_String|None = None) -> b.Expression:
        tz = _extract_tz(datetime, tz)
        return _make_expr("datetime_quarter", datetime, tz, b.Int64.ref("res"))

    @classmethod
    def month(cls, datetime: _DateTime, tz: dt.tzinfo|_String|None = None) -> b.Expression:
        tz = _extract_tz(datetime, tz)
        return _make_expr("datetime_month", datetime, tz, b.Int64.ref("res"))

    @classmethod
    def week(cls, datetime: _DateTime, tz: dt.tzinfo|_String|None = None) -> b.Expression:
        tz = _extract_tz(datetime, tz)
        return _make_expr("datetime_week", datetime, tz, b.Int64.ref("res"))

    @classmethod
    def day(cls, datetime: _DateTime, tz: dt.tzinfo|_String|None = None) -> b.Expression:
        tz = _extract_tz(datetime, tz)
        return _make_expr("datetime_day", datetime, tz, b.Int64.ref("res"))

    @classmethod
    def dayofyear(cls, datetime: _DateTime, tz: dt.tzinfo|_String|None = None) -> b.Expression:
        tz = _extract_tz(datetime, tz)
        return _make_expr("datetime_dayofyear", datetime, tz, b.Int64.ref("res"))

    @classmethod
    def hour(cls, datetime: _DateTime, tz: dt.tzinfo|_String|None = None) -> b.Expression:
        tz = _extract_tz(datetime, tz)
        return _make_expr("datetime_hour", datetime, tz, b.Int64.ref("res"))

    @classmethod
    def minute(cls, datetime: _DateTime, tz: dt.tzinfo|_String|None = None) -> b.Expression:
        tz = _extract_tz(datetime, tz)
        return _make_expr("datetime_minute", datetime, tz, b.Int64.ref("res"))

    @classmethod
    def second(cls, datetime: _DateTime) -> b.Expression:
        return _make_expr("datetime_second", datetime, b.Int64.ref("res"))

    @classmethod
    def isoweekday(cls, datetime: _DateTime, tz: dt.tzinfo|_String|None = None) -> b.Expression:
        """
        Return the ISO weekday as an integer, where Monday is 1, and Sunday is 7.
        """
        tz = _extract_tz(datetime, tz)
        return _make_expr("datetime_weekday", datetime, tz, b.Int64.ref("res"))

    @classmethod
    def weekday(cls, datetime: _DateTime, tz: dt.tzinfo|_String|None = None) -> b.Expression:
        return cls.isoweekday(datetime, tz) - 1 # Convert ISO weekday (1=Mon..7=Sun) to weekday (0=Mon..6=Sun)

    @classmethod
    def fromordinal(cls, ordinal: _Integer) -> b.Expression:
        # Convert ordinal to milliseconds, since ordinals in Python are days
        # Minus 1 day since we can't declare date 0000-00-00
        ordinal_milliseconds = (ordinal - 1) * 86400000 # 24 * 60 * 60 * 1000
        return cls.add(b.DateTime(dt.datetime(1, 1, 1, 0, 0, 0)), milliseconds(ordinal_milliseconds))

    @classmethod
    def strptime(cls, date_str: _String, format: _String) -> b.Expression:
        return _make_expr("parse_datetime", date_str, format, b.DateTime.ref("res"))

    @classmethod
    def to_date(cls, datetime: _DateTime, tz: dt.tzinfo|_String|None = None) -> b.Expression:
        tz = _extract_tz(datetime, tz)
        return _make_expr("construct_date_from_datetime", datetime, tz, b.Date.ref("res"))

    @classmethod
    def format(cls, date: _DateTime, format: _String, tz: dt.tzinfo|_String|None = None) -> b.Expression:
        tz = _extract_tz(date, tz)
        return _make_expr("datetime_format", date, format, tz, b.String.ref("res"))

    @classmethod
    def add(cls, date: _DateTime, period: b.Producer) -> b.Expression:
        return _make_expr("datetime_add", date, period, b.DateTime.ref("res"))

    @classmethod
    def subtract(cls, date: _DateTime, period: b.Producer) -> b.Expression:
        return _make_expr("datetime_subtract", date, period, b.DateTime.ref("res"))

    @classmethod
    def range(cls, start: _DateTime | None = None, end: _DateTime | None = None, periods: int = 1, freq: Frequency = "D") -> b.Expression:
        """
        Note on date_ranges and datetime_range: The way the computation works is that it first overapproximates the
        number of periods.

        For example, date_range(2025-02-01, 2025-03-01, freq='M') and date_range(2025-02-01, 2025-03-31, freq='M') will
        compute range_end to be ceil(28*1/(365/12))=1 and ceil(58*1/(365/12))=2.

        Then, the computation fetches range_end+1 items into _date, which is the right number in the first case but
        one too many in the second case. That's why a filter end >= _date (or variant of) is applied, to remove any
        extra item. The result is two items in both cases.
        """
        if start is None and end is None:
            raise ValueError("Invalid start/end datetime for datetime_range. Must provide at least start datetime or end datetime")
        _milliseconds = {
            "ms": 1,
            "s": 1 / 1_000,
            "m": 1 / 60_000,
            "H": 1 / 3_600_000,
            "D": 1 / 86_400_000,
            "W": 1 / (86_400_000 * 7),
            "M": 1 / (86_400_000 * (365 / 12)),
            "Y": 1 / (86_400_000 * 365),
        }
        date_func = cls.add
        if start is None:
            start = end
            end = None
            date_func = cls.subtract
        assert start is not None
        if end is not None:
            num_ms = cls.period_milliseconds(start, end)
            if freq == "ms":
                _end = num_ms
            else:
                _end = std.cast(b.Int64, std.math.ceil(num_ms * _milliseconds[freq]))
            # datetime_range is inclusive. add 1 since std.range is exclusive
            ix = std.range(0, _end + 1, 1)
        else:
            ix = std.range(0, periods, 1)
        _date = date_func(start, _periods[freq](ix))
        if isinstance(end, dt.datetime) :
            return b.DateTime(end) >= _date
        elif end is not None:
            return end >= _date
        return _date

    @classmethod
    def period_milliseconds(cls, start: _DateTime, end: _DateTime) -> b.Expression:
        return _make_expr("datetimes_period_milliseconds", start, end, b.Int64.ref("res"))

#--------------------------------------------------
# Periods
#--------------------------------------------------

def nanoseconds(period: _Integer) -> b.Expression:
    return _make_expr("nanosecond", std.cast_to_int64(period), b.Int64.ref("res"))

def microseconds(period: _Integer) -> b.Expression:
    return _make_expr("microsecond", std.cast_to_int64(period), b.Int64.ref("res"))

def milliseconds(period: _Integer) -> b.Expression:
    return _make_expr("millisecond", std.cast_to_int64(period), b.Int64.ref("res"))

def seconds(period: _Integer) -> b.Expression:
    return _make_expr("second", std.cast_to_int64(period), b.Int64.ref("res"))

def minutes(period: _Integer) -> b.Expression:
    return _make_expr("minute", std.cast_to_int64(period), b.Int64.ref("res"))

def hours(period: _Integer) -> b.Expression:
    return _make_expr("hour", std.cast_to_int64(period), b.Int64.ref("res"))

def days(period: _Integer) -> b.Expression:
    return _make_expr("day", std.cast_to_int64(period), b.Int64.ref("res"))

def weeks(period: _Integer) -> b.Expression:
    return _make_expr("week", std.cast_to_int64(period), b.Int64.ref("res"))

def months(period: _Integer) -> b.Expression:
    return _make_expr("month", std.cast_to_int64(period), b.Int64.ref("res"))

def years(period: _Integer) -> b.Expression:
    return _make_expr("year", std.cast_to_int64(period), b.Int64.ref("res"))


Frequency = Union[
    Literal["ms"],
    Literal["s"],
    Literal["m"],
    Literal["H"],
    Literal["D"],
    Literal["W"],
    Literal["M"],
    Literal["Y"],
]

_periods = {
    "ms": milliseconds,
    "s": seconds,
    "m": minutes,
    "H": hours,
    "D": days,
    "W": weeks,
    "M": months,
    "Y": years,
}

def _extract_tz(datetime: _DateTime, tz: dt.tzinfo|_String|None) -> _String:
    default_tz = "UTC"
    if tz is None:
        if isinstance(datetime, dt.datetime):
            tz = datetime.tzname() or default_tz
        else:
            tz = default_tz
    elif isinstance(tz, dt.tzinfo) :
        tz = tz.tzname(None) or default_tz
    return tz



"""
RelationalAI Multi-Connection Configuration System

This package contains the new Pydantic-based configuration system with:
- Multi-connection support (Snowflake, DuckDB)
- Multi-profile support (dev, prod, staging, etc.)
- Discriminated unions for type-safe connection handling
- Auto-loading from multiple sources (raiconfig.yaml, Snowflake connections, DBT profiles)
- Environment variable templating support (YAML)
- Profile override system (profile values override config defaults)
- Type-safe validation via Pydantic
"""

from .config import Config, _Config, RAIConfig, ConfigFromDBT, ConfigFromSnowflake
from .config_fields import EngineConfig, DataConfig, CompilerConfig, ModelConfig, ReasonerConfig, DebugConfig
from .connections import (
    ConnectionConfig,
    SnowflakeConnection,
    DuckDBConnection,
    UsernamePasswordAuth,
    UsernamePasswordMFAAuth,
    ExternalBrowserAuth,
    JWTAuth,
    OAuthAuth,
    ProgrammaticAccessTokenAuth,
)

__all__ = [
    # Main classes
    "Config",
    "_Config",
    "RAIConfig",
    "ConfigFromDBT",
    "ConfigFromSnowflake",

    # Nested config models
    "EngineConfig",
    "DataConfig",
    "CompilerConfig",
    "ModelConfig",
    "ReasonerConfig",
    "DebugConfig",

    # Connection types
    "ConnectionConfig",
    "SnowflakeConnection",
    "DuckDBConnection",

    # Snowflake authenticators
    "UsernamePasswordAuth",
    "UsernamePasswordMFAAuth",
    "ExternalBrowserAuth",
    "JWTAuth",
    "OAuthAuth",
    "ProgrammaticAccessTokenAuth",
]



from v0.relationalai.early_access.dsl.core.types import Type
from v0.relationalai.early_access.dsl.core.utils import camel_to_snake, to_rai_way_string, to_pascal_case
from v0.relationalai.early_access.dsl.constants import VERBAL_PART_CONNECTION
from v0.relationalai.early_access.dsl.ontologies.roles import Role

def normalize(s):
    return "\n".join(line.strip() for line in s.splitlines() if line.strip())

def build_relation_name(roles, verbal_parts):
    """
    Build the relation name from the verbal parts and roles. The relation name does not include the namespace.
    We can derive the name of the relation from the Role names and the reading (verbal_parts).

    Simple case: the Relationship is either unary or binary.
    In case when the reading starts with:
    - "is": we use the reading omitting "is";
    - "has": we use the reading omitting "has";
    if role doesn't have a name then role's player name transformed from camel case to snake is used.
    - otherwise, the reading is used.
    Examples:
    - reading: "{Person} is female" -> namespace: "person", relation name: "female"
    - reading: "{Person} has birth {Date}" -> namespace: "person", relation name: "birth"
    - reading: "{Account} has {AccountNr}" -> namespace: "account", relation name: "nr" (or "account_nr" when role has no name)
    - reading: "{Account} created on {Date}" -> namespace: "account", relation name: "created_on"

    Complex case: the Relationship has more than two Roles.
    We typically omit the name of the first Role in the reading, use the rest of the reading filling in
    the role name positions with the names.
    Examples:
    - reading: "{Person} has {Child} born on {Date}" -> namespace: "person", relation name: "has_child_born_on_date"
    - reading: "{Account} on {Date} has activation- {Event}" -> namespace: "account", relation name: "on_date_has_activation_event"

    """
    rel_name = ""
    if len(roles) == 1:  # for unary relations
        if len(verbal_parts) > 0:
            rel_name = to_rai_way_string(verbal_parts[0])
        if not rel_name:
            rel_name = _get_role_name(roles[0])
    elif len(roles) == 2:  # for binary relations
        s_role = roles[1]
        if len(verbal_parts) > 0:
            rel_name = to_rai_way_string(verbal_parts[0])
            if verbal_parts[0].strip().startswith(("has", "is")):
                if not rel_name:
                    rel_name += _get_role_name(s_role)
        else:
            rel_name = _get_role_name(s_role)
    else:
        # assumption that len(roles) - 1 == len(verbal_parts)
        join_parts = [to_rai_way_string(verbal_parts[0], False)]
        for i in range(1, len(roles)): # skip the first role
            role = roles[i]
            join_parts.append(_get_role_name(role))
            if i < len(roles) - 1:  # append verbal part after each iteration except the last one
                join_parts.append(to_rai_way_string(verbal_parts[i]))
        rel_name = VERBAL_PART_CONNECTION.join(join_parts)
    return rel_name

def build_relationship_name(roles, verbal_parts):
    """
    Build the relationship name from the verbal parts and roles.

    """
    relationship_name = ""
    for i in range(len(roles)):
        relationship_name += to_pascal_case(roles[i].verbalize())
        # assumption that len(roles) >= than len(verbal_parts)
        if i < len(verbal_parts):
            relationship_name += to_pascal_case(verbal_parts[i])

    return relationship_name

def extract_relation_text_with_signature(*args):
    types = []
    text_frags = []
    if len(args) < 2:
        raise Exception("Invalid declaration for the unary relation. Type should be followed by a text.")

    for i in range(len(args)):
        a = args[i]
        # even must be 'Type' and odd must be 'str'
        if i % 2 == 0:
            if not isinstance(a, Type):
                raise Exception(f"Invalid declaration for the relation. Argument {i + 1} is not a type.")
            types.append(a)
        else:
            if not isinstance(a, str):
                raise Exception(f"Invalid declaration for the relation. Argument {i + 1} is not a text.")
            text_frags.append(a)
    # check that N-ary relation ends with Type. Exclude the case for unary when args == 2: (Person, "is adult")
    if len(text_frags) == len(types) and len(args) > 2:
        raise Exception("Invalid declaration for the relation. Relations with arity > 2 must end with a type.")

    return text_frags, types

def _get_role_name(role) -> str:
    if isinstance(role, Type):
        return camel_to_snake(role.display())
    elif isinstance(role, Role):
        return to_rai_way_string(role.ref_name(), False) if role.name else camel_to_snake(role.player().display())
    raise Exception(f"Invalid declaration for the relation. Argument {role} is not a 'Type' or a 'Role'.")


"""
API for RelationalAI.
"""

from .internal import (
    Model, Concept, Relationship, RelationshipReading, Expression, Fragment, Error, Field,
    AnyEntity, String, Integer, Int64, Int128, Float, Decimal, Bool,
    Date, DateTime,
    RawSource, Hash,
    select, where, require, define, distinct, union, data,
    rank, asc, desc,
    count, sum, min, max, avg, per,
    not_
)

__all__ = [
    "Model", "Concept", "Relationship", "RelationshipReading", "Expression", "Fragment", "Error", "Field",
    "AnyEntity", "String", "Integer", "Int64", "Int128", "Float", "Decimal", "Bool",
    "Date", "DateTime",
    "RawSource", "Hash",
    "select", "where", "require", "define", "distinct", "union", "data",
    "rank", "asc", "desc",
    "count", "sum", "min", "max", "avg", "per",
    "not_"
]



# type: ignore
import re
from typing import Union, Callable, Tuple, List, Optional, Sequence as PySequence

from v0.relationalai.early_access.dsl import Relation, AssertedRelation
from v0.relationalai.early_access.dsl.core.relations import ExportRelation
from v0.relationalai.early_access.dsl.ontologies.export import Export
from v0.relationalai.early_access.dsl.ontologies.raw_source import RawSource
from v0.relationalai.semantics.metamodel import ir, builtins, types, factory as f
from v0.relationalai.semantics.metamodel.types import Any, Symbol
from v0.relationalai.semantics.metamodel.util import OrderedSet, ordered_set

from v0.relationalai.early_access.dsl.core.constraints.predicate.atomic import Atom
from v0.relationalai.early_access.dsl.core.constraints.scalar import ScalarConstraint
from v0.relationalai.early_access.dsl.core.exprs import Wildcard
from v0.relationalai.early_access.dsl.core.exprs.scalar import Literal, BinaryScalarExpr
from v0.relationalai.early_access.dsl.core.logic import RelVariable
from v0.relationalai.early_access.dsl.core.logic.aggregation import Aggregation
from v0.relationalai.early_access.dsl.core.logic.exists import ExistentialConstraint
from v0.relationalai.early_access.dsl.core.rules import Rule, Annotation
from v0.relationalai.early_access.dsl.core.types import Type
from v0.relationalai.early_access.dsl.core.types.standard import Boolean, Hash, Integer, Decimal, Float, Double, \
    BigInteger, String, Date, DateTime, RowId
from v0.relationalai.early_access.dsl.ontologies.models import Model

class CompilerContext:
    def __init__(self):
        self.var_map:dict[int, ir.Var] = {}
        self.items:OrderedSet[ir.Task] = OrderedSet()
        self.project_away_vars_count:int = -1
        self.project_away_vars:OrderedSet[ir.Var] = OrderedSet()

    def to_var(self, item:Union[RelVariable, Literal], t:Optional[ir.Type] = None) -> ir.Var:
        entity_id  = item.entityid()
        if entity_id not in self.var_map:
            if isinstance(item, RelVariable):
                self.var_map[entity_id] = f.var(item.display(), t)
            else:
                self.var_map[entity_id] = f.lit(item.val)
        return self.var_map[entity_id]

    def project_away_var_name(self) -> str:
        self.project_away_vars_count += 1
        return f"_pa_{self.project_away_vars_count}"

    def to_project_away_var(self, t: ir.Type) -> ir.Var:
        var = f.var(self.project_away_var_name(), t)
        self.project_away_vars.add(var)
        return var

    def add(self, item:ir.Task):
        self.items.add(item)

    def clone(self):
        c = CompilerContext()
        c.var_map = self.var_map.copy()
        return c

class Compiler:
    def __init__(self):
        self.types: dict[Type, ir.Type] = {
                Hash: f.scalar_type("UInt128"),
                Boolean: types.Bool,
                Integer: types.Int128,
                Decimal: types.Decimal,
                Float: types.Number,
                Double: types.Number,
                String: types.String,
                BigInteger: f.scalar_type("Int128"),
                Date: f.scalar_type("Date"),
                DateTime: f.scalar_type("DateTime"),
                Symbol: types.Symbol,
                RowId: types.Sha1,
                Any: types.Any
        }
        self.relations: dict[int, ir.Relation] = {}

    def to_type(self, t: Type) -> ir.Type:
        if t not in self.types:
            if t._name in types.builtin_scalar_types_by_name:
                self.types[t] = types.builtin_scalar_types_by_name[t._name]
            else:
                self.types[t] = f.scalar_type(t._name)
        return self.types[t]

    @staticmethod
    def _parse_relation_name(name: str, producer: Callable[[str, Any], Union[ir.Field, ir.Var]]) \
            -> Tuple[str, List[Union[ir.Field, ir.Var]]]:
        parts = re.split(r'(?<!:):(?!:)', name)
        if len(parts) > 1:
            var_list = [producer(f":{part}", types.Symbol) for part in parts[1:]]
            return parts[0], var_list
        return name, []

    def to_relation(self, ctx: CompilerContext, item:Atom, name:str) -> ir.Relation:
        # We use `item.relcomp.entityid()` instead of `item.entityid()`
        # to ensure that there is only one instance of a relation instead of multiple instances.
        #
        # Example:
        # - `test(x in Int, y in DateTime)`
        # - `test(y in Int, x in DateTime)`
        #
        # These will be considered the same relation.
        entity_id = item.relcomp.entityid()
        if entity_id not in self.relations:
            if name in builtins.builtin_relations_by_name:
                self.relations[entity_id] = builtins.builtin_relations_by_name[name]
            else:
                name, fields = self._parse_relation_name(name, f.field)
                fields.extend(
                    f.field(
                        arg.display() if not isinstance(arg, Wildcard) else ctx.project_away_var_name(),
                        self.to_type(role.root_unconstrained_type())
                    )
                    for arg, role in zip(item.args, item.relcomp._signature._types)
                )
                annotations = [builtins.external_annotation] if isinstance(item.relcomp, AssertedRelation) else []
                self.relations[entity_id] = f.relation(name, fields, annos=annotations)
        return self.relations[entity_id]

    def to_args(self, ctx: CompilerContext, item: Atom) -> tuple[list[ir.Value], list[ir.Value]]:
        args_list = []
        project_away_vars = []

        for arg, role in zip(item.args, item.relcomp._signature._types):
            if isinstance(arg, Wildcard):
                project_away_var = ctx.to_project_away_var(self.to_type(role.root_unconstrained_type()))
                project_away_vars.append(project_away_var)
                args_list.append(project_away_var)
            else:
                args_list.append(self.to_var(ctx, arg))

        return args_list, project_away_vars

    def to_vars(self, ctx:CompilerContext, item:ExistentialConstraint) -> set[ir.Var]:
        return {self.to_var(ctx, cur) for cur in item._scalars.values()}

    def to_var(self, ctx:CompilerContext, item:Union[RelVariable, Literal]) -> ir.Var:
        return ctx.to_var(item, self.to_type(item._type.root_unconstrained_type())) \
            if isinstance(item, RelVariable) else ctx.to_var(item)

    def compile_head(self, ctx:CompilerContext, head:Atom, original_name:str,
                     annotations: Optional[PySequence[ir.Annotation]]=None) -> ir.Task:
        name, sig_vars = self._parse_relation_name(original_name, f.var)
        sig_vars.extend(self.to_var(ctx, cur) for cur in head.args)
        return self._compile_output(head, sig_vars) if name == 'output' \
            else f.derive(self.to_relation(ctx, head, original_name), sig_vars, annotations)

    @staticmethod
    def _compile_output(item:Atom, sig_vars: List[Union[ir.Field, ir.Var]]):
        annotations = [builtins.export_annotation, builtins.external_annotation] \
            if isinstance(item.relcomp, ExportRelation) else []
        return f.output([(var.name, var) for var in sig_vars], annos=annotations)

    @staticmethod
    def _compile_annotations(annotations:PySequence[Annotation]) -> Optional[PySequence[ir.Annotation]]:
        return [f.annotation(f.relation(str(a), []), []) for a in annotations]

    def compile_atom(self, ctx:CompilerContext, atom:Atom) -> Union[ir.Construct, ir.Lookup, ir.Not]:
        original_name = atom.relcomp.qualified_name()
        name, args = self._parse_relation_name(original_name, f.var)
        if name.startswith('^'):
            vs = tuple(self.to_var(ctx, arg) for arg in atom.args[:-1])
            return ir.Construct(None, (name[1:], *vs), self.to_var(ctx, atom.args[-1]))
        else:
            atom_args, project_away_vars = self.to_args(ctx, atom)
            args.extend(atom_args)
            task = f.lookup(self.to_relation(ctx, atom, original_name), args)
            if len(project_away_vars) > 0:
                task = f.exists(list(project_away_vars), f.logical([task]))
            return f.not_(task) if atom.negated else task

    def _compile_scalar_expressions(self, ctx, scalar_constraint):
        left = scalar_constraint.left
        right = scalar_constraint.right

        expr = right if isinstance(right, (BinaryScalarExpr, Aggregation)) else left
        other = left if expr is right else right

        return (
            self._compile_binary_scalar_expressions(ctx, expr, other)
            if isinstance(expr, BinaryScalarExpr)
            else self._compile_aggregation(ctx, expr, other)
        )

    # Example of a binary scalar expression: ( c = a / b ) or ( a / b = c )
    def _compile_binary_scalar_expressions(self, ctx, left:BinaryScalarExpr, right:RelVariable):
        args = [self.to_var(ctx, left._left), self.to_var(ctx, left._right), self.to_var(ctx, right)]
        return f.lookup(builtins.builtin_relations_by_name[left.op()], args)

    # Example of an aggregation:
    #   - ( d = sum[(e): l1(c, e) and l2(a, e) ] ) or
    #   - ( sum[(e): l1(c, e) and l2(a, e) ] = d )
    # Expected IR:
    #   Logical
    #     Lookup l1(c, e)
    #     Lookup l2(a, e)
    #     Aggregate sum( [], [a, c], [e, d] )
    #     Hoist d a c
    #
    # We need to handle `count` aggregation differently.
    #   - ( d = count[(e): l1(e) ] )
    # Expected IR:
    #   Logical
    #     Lookup l1(e)
    #     Aggregate count( [e], [], [d] )
    #     Hoist d
    def _compile_aggregation(self, ctx, left:Aggregation, right:RelVariable) -> ir.Aggregate:
        method = left._method
        result_var = self.to_var(ctx, right)
        vargs = [self.to_var(ctx, var) for var in left._scalars.values()]
        items, lookup_args = self._compile_atoms_and_scalar_constraints(ctx, left._atoms.values(), left._sconstraints.values())
        group_vars = lookup_args - vargs
        agg_relation = builtins.builtin_relations_by_name[method]
        projection = vargs if self._is_count_aggregation(method) else vargs[:-1]
        args = [result_var] if self._is_count_aggregation(method) else [vargs[-1], result_var]
        items.add(f.aggregate(agg_relation, projection, group_vars.list, args))
        return f.logical(list(items), [result_var, *group_vars.list])

    def compile_scalar_constraint(self, ctx:CompilerContext, scalar_constraint:ScalarConstraint) -> ir.Task:
        if all(isinstance(side, BinaryScalarExpr) for side in (scalar_constraint.left, scalar_constraint.right)):
            raise Exception("It's not allowed for left and right operands to be BinaryScalarExpr.")
        if any(isinstance(side, (BinaryScalarExpr, Aggregation)) for side in (scalar_constraint.right, scalar_constraint.left)):
            return self._compile_scalar_expressions(ctx, scalar_constraint)
        else:
            args = [self.to_var(ctx, scalar_constraint.left), self.to_var(ctx, scalar_constraint.right)]
            return f.lookup(builtins.builtin_relations_by_name[scalar_constraint.op], args)

    # TODO: Replace to this one when we implement this in the to Rel Emitter
    # def compile_existential(self, ctx:CompilerContext, existential:ExistentialConstraint) -> ir.Task:
    #     items, vargs = (self._compile_atoms_and_scalar_constraints(ctx, existential._atoms.values(), existential._sconstraints.values()))
    #     return f.logical(list(items), (vargs - self.to_vars(ctx, existential)).list)

    def compile_existential(self, ctx:CompilerContext, existential:ExistentialConstraint) -> ir.Task:
        items = ordered_set()
        for atom in existential._atoms.values():
            items.add(self.compile_atom(ctx, atom))
        for scalar_constraint in existential._sconstraints.values():
            items.add(self.compile_scalar_constraint(ctx, scalar_constraint))
        return f.exists(list(self.to_vars(ctx, existential)), f.logical(list(items)))

    def _compile_atoms_and_scalar_constraints(self, ctx: CompilerContext, atoms: [Atom], scalar_constraints: [ScalarConstraint]) \
            -> Tuple[OrderedSet[ir.Task], OrderedSet[ir.Var]]:
        items = ordered_set()
        vargs = ordered_set()
        for atom in atoms:
            comp_atom = self.compile_atom(ctx, atom)
            if not isinstance(comp_atom, ir.Construct):
                vargs.update(arg for arg in comp_atom.args if self._arg_filter(arg))
            items.add(comp_atom)
        for scalar_constraint in scalar_constraints:
            comp_scalar = self.compile_scalar_constraint(ctx, scalar_constraint)
            if isinstance(comp_scalar, ir.Lookup):
                vargs.update(arg for arg in comp_scalar.args if self._arg_filter(arg))
            elif isinstance(comp_scalar, ir.Logical):
                vargs.update(comp_scalar.hoisted)
            items.add(comp_scalar)
        return items, vargs

    @staticmethod
    def _arg_filter(arg):
        return not isinstance(arg, ir.Literal) and not arg.name.startswith(":")

    @staticmethod
    def _is_count_aggregation(method: str) -> bool:
        return method == "count"

    def compile_rule(self, rule:Rule, name:str) -> ir.Task:
        ctx = CompilerContext()
        for existential in rule._existentials.values():
            if len(existential._scalars) != 0:
                ctx.add(self.compile_existential(ctx, existential))
            else:
                for atom in existential._atoms.values():
                    ctx.add(self.compile_atom(ctx, atom))
                for scalar_constraint in existential._sconstraints.values():
                    ctx.add(self.compile_scalar_constraint(ctx, scalar_constraint))
        for atom in rule._atoms.values():
            ctx.add(self.compile_atom(ctx, atom))
        for scalar_constraint in rule._sconstraints.values():
            ctx.add(self.compile_scalar_constraint(ctx, scalar_constraint))
        ctx.add(self.compile_head(ctx, rule.head, name, self._compile_annotations(rule._annotations)))
        return f.logical(list(ctx.items))

    def compile_raw_sources(self, raw_sources: OrderedSet[RawSource]) -> ir.Model:
        return f.compute_model(
            f.logical([
                f.logical([f.derive(builtins.raw_source, [rs.language, rs.raw_source])])
                for rs in raw_sources
            ])
        )

    def compile_model(self, model: Model) -> ir.Model:
        return f.compute_model(
            f.logical(
                [self.compile_rule(rule, relation.qualified_name())
                 for relation in model._relations.values()
                 for rule in relation._rules]
            )
        )

    def compile_queries(self, queries: OrderedSet[Relation]) -> ir.Model:
        return f.compute_model(
            f.logical([
                self.compile_rule(rule, query.qualified_name())
                for query in queries
                for rule in query._rules
            ])
        )

    def compile_export(self, export: Export) -> ir.Model:
        return f.compute_model(
            f.logical([
                self.compile_rule(rule, column.qualified_name())
                for column in export.columns
                for rule in column._rules
            ])
        )



import warnings

from v0.relationalai.semantics.metamodel.factory import (compute_model, derive, field, lit, logical, output, relation,
                                                      property, var, union, lookup, Not, exists, annotation, scalar_type,
                                                      aggregate, entity, construct, literal, input_field, engine, success,
                                                      decimal_type, model)

__all__ = ["compute_model", "derive", "field", "lit", "logical", "output", "relation", "property", "var", "union",
           "lookup", "Not", "exists", "annotation", "scalar_type", "aggregate", "entity", "construct", "literal",
           "input_field", "engine", "success", "decimal_type", "model"]

warnings.warn(
    "relationalai.early_access.metamodel.factory is deprecated, "
    "Please migrate to relationalai.semantics.metamodel.factory",
    DeprecationWarning,
    stacklevel=2,
)